package com.ef_prime.rflow;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;

import javax.swing.JPanel;
import javax.swing.border.Border;
import javax.swing.border.LineBorder;
import javax.swing.text.BadLocationException;

import com.ef_prime.rflow.exception.IncompatibleCommandException;
import com.ef_prime.rflow.flow.model.FlowNode;
import com.ef_prime.rflow.node.NodeModel;
import com.ef_prime.rflow.node.NodeModelChangeListener;
import com.ef_prime.rflow.tool.CodingTextPane;

import edu.uci.ics.jung.visualization.PickedState;

public class CodeBar extends JPanel {
	
	//JLabel label = new JLabel(" ");
	//JTextField codeField = new JTextField(" "); 
	private CodingTextPane codeField = new CodingTextPane(); 
	private FlowNode node;
	
	public CodeBar(){
		super();
		setLayout(new BorderLayout());
		//setBorder(new EmptyBorder(2,10,2,10));
		//setBorder(new EmptyBorder(0,0,0,0));
		final Border focusBorder = new LineBorder(Color.GRAY, 1);
		final Border noFocusBorder = new LineBorder(Color.LIGHT_GRAY, 1);
		setBorder(noFocusBorder);

		setPreferredSize(new Dimension(Integer.MAX_VALUE, 25));
		setMaximumSize(new Dimension(Integer.MAX_VALUE, 25));
		setMinimumSize(new Dimension(0, 25));
		
		
		codeField.setIntend(false);
		codeField.setFont(Prefs.CODE_FONT_12);
		codeField.addFocusListener(new FocusListener(){
			public void focusGained(FocusEvent e) {
				CodeBar.this.setBorder(focusBorder);
			}
			
			public void focusLost(FocusEvent e) {
				CodeBar.this.setBorder(noFocusBorder);
				sendCode2Model();
			}
		});
		codeField.addKeyListener(new KeyAdapter(){
			@Override
			public void keyPressed(KeyEvent e) {
				if(e.isConsumed())return;
				switch(e.getKeyCode()){
				case KeyEvent.VK_ENTER:
					//sendCode2Model();
					CodeBar.this.requestFocus();
					e.consume();
					break;
				}
			}
		});
		
		add(codeField);
	}
	
	public void clear(){
		try {
			codeField.getDocument().remove(0, codeField.getDocument().getLength());
		} catch (BadLocationException e) {}
		
		codeField.setEditable(false);
	}
	
	public void setText(String text){
		if(text == null || text.length() < 1)
			codeField.setText(" "); //$NON-NLS-1$
		else
			codeField.setText(text);
		
		codeField.setCaretPosition(0);
		codeField.getUndoManager().discardAllEdits();
	}
	
	private void sendCode2Model(){
		NodeModel model = (node == null)?null:node.getModel();
		if(model != null && model.isEditable()){
			try {
				NodeModel newModel = model.copy();
				newModel.setCommand(codeField.getRawText());
				node.setModel(newModel);
				requestFocus();
			} catch (IncompatibleCommandException e1) {
				codeField.setSelectionStart(0);
				codeField.setSelectionEnd(codeField.getDocument().getLength());
				codeField.requestFocus();
			}
		}
	}
	
	private void setModel(FlowNode node){
		this.node = node;
		NodeModel model = (node == null)?null:node.getModel();
		if(model != null){
			codeField.setEditable(model.isEditable());
			if(!model.isEditable()){
				codeField.setBackground(codeField.getParent().getBackground());
			}
			String c = model.getCommand();
			setText((model.isEditable())?c:omitCommand(c));
		}else{
			codeField.setEditable(false);
			codeField.setBackground(codeField.getParent().getBackground());
			//codeField.setEnabled(false);
			setText(null);
		}
	} 
	
	private String omitCommand(String command){
		String omitted = command.replaceAll(Constants.NEWLINE_STR, "  "); //$NON-NLS-1$
		omitted = omitted.replaceAll("\t", ""); //$NON-NLS-1$ //$NON-NLS-2$
		if(omitted.length() > 50)
			omitted = omitted.substring(0, 50)+"..."; //$NON-NLS-1$
		return omitted;
	}
	
	final ItemListener pickedStateItemListener = new ItemListener(){
		public void itemStateChanged(ItemEvent e) {
			PickedState pickedState = (PickedState)e.getSource();
			if(pickedState.getPickedVertices().size() == 1){
				FlowNode node = (FlowNode)pickedState.getPickedVertices().iterator().next();
				setModel(node);
			}else{
				setModel(null);
			}
		}
	};
	
	final NodeModelChangeListener modelChangeListener = new NodeModelChangeListener(){
		public void nodeModelChanged(FlowNode node, NodeModel old, NodeModel model) {
			if(node == CodeBar.this.node)
				setModel(node);
		}
	};
}
