package com.ef_prime.rflow;


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Locale;

import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

import com.ef_prime.rflow.Setting.Key;
import com.ef_prime.rflow.resource.Messages;
import com.ef_prime.rflow.util.FileUtility;
import com.ef_prime.rflow.util.PropertiesUtility;

public class Main {

	public static void main(String[] args)
	{
		File runScriptFile = new File(Platform.getUserDir(),"run.vbs"); //$NON-NLS-1$
		File initR = new File(Platform.getUserDir(),"init.R"); //$NON-NLS-1$
		File settingsFile = FileResource.SETTING_FILE;
		File defaultSettingsFile = new File("rflow.settings.default");		 //$NON-NLS-1$
		initR.delete();

		if(!settingsFile.exists() && defaultSettingsFile.exists())
			try {
				FileUtility.copy(defaultSettingsFile, settingsFile);
			} catch (IOException e) {
			}

		final Setting setting = Setting.defaultSetting();

		try
		{
			PropertiesUtility.load(setting, settingsFile);
		}
		catch(Exception ex)
		{
			//設定ファイルが読み込めないエラー
			System.out.println("DEBUG:設定ファイルが読み込めないエラー"); //$NON-NLS-1$
			System.out.println("DEBUG:Cannot read:" + settingsFile.getAbsolutePath()); //$NON-NLS-1$
			showSettingDialog();
			return;
		}

		//---------------------------------
		//        Rのバージョンの取得
		//---------------------------------

		String r_home = (String)setting.getProperty(Key.R_HOME);
		boolean use_library_pack = (Boolean)setting.getProperty(Key.USE_LIBLARY_PACK);
		Locale.setDefault(new Locale(((String)setting.getProperty(Key.LANGUAGE))));

		
		File rHome = new File(r_home);

		if(!rHome.exists())
		{
			System.out.println("DEBUG:Directory R_HOME not found." + r_home); //$NON-NLS-1$
			showSettingDialog();
			return;
		}

		File rBin = new File(rHome,"bin"); //$NON-NLS-1$
		//File rExe = new File(rBin,"R.exe");
		File rScriptExe = new File(rBin,"Rscript.exe"); //$NON-NLS-1$
		
		if(!rScriptExe.exists())
		{
			System.out.println("DEBUG:r.exe not found"); //$NON-NLS-1$
			showSettingDialog();
			return;
		}
		
		String version,libpath;
		
		String command = "\"" + rScriptExe.getAbsolutePath() + "\" r_library_path.R"; //$NON-NLS-1$ //$NON-NLS-2$
		System.out.println(command);
		try {
			Process p = Runtime.getRuntime().exec(command);
			BufferedReader reader = new BufferedReader(
					new InputStreamReader(p.getInputStream()));
			String major = reader.readLine();
			String minor = reader.readLine();
			version = major + "." + minor; //$NON-NLS-1$
			libpath = reader.readLine();

			reader.close();
		} catch (IOException e) {
			System.out.println("DEBUG:Error at " + command); //$NON-NLS-1$
			showSettingDialog();
			return;
		}
		//該当ライブラリパックが存在
		File libraryPack = new File(new File("LibraryPack"),version); //$NON-NLS-1$

		if(use_library_pack && libraryPack.exists())
		{
			//R起動時に渡すスクリプトを生成
			try {
				BufferedWriter writer = new BufferedWriter(new FileWriter(initR));
				String rlib = Constants.replaceFileSeparator(libraryPack.getCanonicalPath());
				writer.write("library(\"rJava\", lib.loc=\""+ rlib + "\")"); //$NON-NLS-1$ //$NON-NLS-2$
				writer.newLine();
				writer.write("library(\"JavaGD\", lib.loc=\""+ rlib + "\")"); //$NON-NLS-1$ //$NON-NLS-2$
				writer.newLine();
				writer.flush();
				writer.close();
			} catch (IOException e1) {
				//TODO エラー処理
			}
			
			
			
			File jriJar = new File(libraryPack,"rJava\\jri\\JRI.jar"); //$NON-NLS-1$
			try {
				makeRunScript(runScriptFile,
						rHome.getAbsolutePath(),
						rBin.getAbsolutePath(),
						libraryPack.getAbsolutePath(),
						jriJar.getAbsolutePath());
			} catch (IOException e) {
				// TODO Auto-generated catch block
			}
		}
		else
		{
			final File jriJar = new File(libpath,"rJava\\jri\\JRI.jar"); //$NON-NLS-1$
			
			if(!jriJar.exists())
			{
				SwingUtilities.invokeLater(new Runnable(){
					public void run() {
						new JPanel();
				        try {
							UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
						} catch (Exception e) {}
						JOptionPane.showMessageDialog(null, Messages.getString("Error.file.file_not_found") + jriJar.getAbsolutePath()); //$NON-NLS-1$
					}});
				return;
			}
			
			try {
				makeRunScript(runScriptFile,
						rHome.getAbsolutePath(),
						rBin.getAbsolutePath(),
						jriJar.getParentFile().getAbsolutePath(),
						jriJar.getAbsolutePath());
			} catch (IOException e) {
				System.out.println("DEBUG:File not exists:" + jriJar); //$NON-NLS-1$
				showSettingDialog();
				return;
			}
		}

		String runCommand = "cscript \"" + runScriptFile.getAbsolutePath() + "\"" + ((args.length > 0)?(" \"" + args[0] + '"'):""); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
		try {
			//Process p = 
				Runtime.getRuntime().exec(runCommand);
		} catch (IOException e) {
		}
	}
	
	static void showSettingDialog()
	{
		SwingUtilities.invokeLater(new Runnable(){
			public void run() {
				//setLookAndFeel
				new JPanel();
		        try {
					UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
				} catch (Exception e) {}
				int val = JOptionPane.showConfirmDialog(null, Messages.getString("Error.Main.error_ask_show_pref")); //$NON-NLS-1$
				if(val == JOptionPane.YES_OPTION)
				{
					SettingDialog.main(new String[0]);
				}
			}});
	}
	
	
	static void makeRunScript(File runScript,String rHome,String rbinDirectory,String libpath,String jriJar)throws IOException{

		BufferedWriter writer = new BufferedWriter(new FileWriter(runScript)); //$NON-NLS-1$

		writer.write("Set objShell = WScript.CreateObject(\"WScript.Shell\")"); //$NON-NLS-1$
		writer.newLine();
		writer.write("Set WSHArguments = WScript.Arguments"); //$NON-NLS-1$
		writer.newLine();
		writer.write("If WSHArguments.Count > 0 Then"); //$NON-NLS-1$
		writer.newLine();
		writer.write("	For Each I In WSHArguments"); //$NON-NLS-1$
		writer.newLine();
		writer.write("		str=str & \" \"\"\" & I & \"\"\"\""); //$NON-NLS-1$
		writer.newLine();
		writer.write("	Next"); //$NON-NLS-1$
		writer.newLine();
		writer.write("End If"); //$NON-NLS-1$
		writer.newLine();
		writer.write("Set colEnv = objShell.Environment(\"Process\")"); //$NON-NLS-1$
		writer.newLine();
		writer.write("colEnv.Item(\"PATH\") = colEnv.Item(\"PATH\") & \";" + rbinDirectory + ";" + libpath + "\""); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		writer.newLine();
		writer.write("colEnv.Item(\"R_HOME\") = \"" + rHome + "\""); //$NON-NLS-1$ //$NON-NLS-2$
		writer.newLine();
		writer.write("library = \"-Djava.library.path=\"\"" + libpath + "\"\"\""); //$NON-NLS-1$ //$NON-NLS-2$
		writer.newLine();
		writer.write("jrijar = \"" + jriJar + "\""); //$NON-NLS-1$ //$NON-NLS-2$
		writer.newLine();
		writer.write("objShell.Run \"javaw.exe -splash:img\\R-flow-splash.png \" & library & \" -classpath .\\lib\\commons-collections-3.1.jar;.\\lib\\colt.jar;.\\lib\\resolver.jar;.\\lib\\jung-1.7.6.jar;.\\lib\\concurrent.jar;.\\lib\\xercesImpl.jar;.\\lib\\xml-apis.jar;.\\lib\\rflow.jar;.\\bin;\"\"\" & jrijar & \"\"\"; com.ef_prime.rflow.Rflow\" & str"); //$NON-NLS-1$
		writer.newLine();
		writer.flush();
		writer.close();
	}
}
/*
#Tue Feb 26 11:42:02 JST 2008
USE_LIBLARY_PACK=true
MAX_PATH=255
IS_FIRST=false
GRID_DISTANCE=10
IS_AUTO_UPDATE=false
USE_CONSOLE_BUFFER=true
R_HOME=C\:\\R\\R-2.6.2\\bin
LANGUAGE=ja
CHECKED_VERSION=0.3.0
USE_GRID=true
DEFAULT_WORKING_DIRECTORY=C\:\\Documents and Settings\\nagai\\My Documents\\eclipse\\Rflow\\data
 */