package com.ef_prime.rflow;

import java.awt.Component;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultListCellRenderer;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.UIManager;
import javax.swing.border.EmptyBorder;
import javax.swing.border.TitledBorder;
import javax.swing.text.JTextComponent;

import com.ef_prime.rflow.Setting.Key;
import com.ef_prime.rflow.resource.Messages;
import com.ef_prime.rflow.tool.ButtonPanel;
import com.ef_prime.rflow.util.FileChooserUtility;
import com.ef_prime.rflow.util.PropertiesUtility;

public class SettingDialog {
	
	private JDialog dialog;
	private Setting setting = null;
	private List<UpdateSetting> updateSettings = new ArrayList<UpdateSetting>();
	
	private SettingDialog(Window parent, final Setting setting, boolean initOnly){
		if(parent instanceof Dialog)
			dialog = new JDialog((Dialog)parent, true);
		else
			dialog = new JDialog((Frame)parent, true);
		
		dialog.setTitle(Messages.getString("Action.preferences")); //$NON-NLS-1$

		
		
		//ボタン
		ActionListener ok = new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				SettingDialog.this.setting = setting.copy();
				if(!update(SettingDialog.this.setting))
					return;
				dialog.dispose();
			}
		};
		ActionListener cancel = new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				SettingDialog.this.setting = null;
				dialog.dispose();
			}
		};
		ButtonPanel buttonPanel = new ButtonPanel(ok, cancel);


		JComponent main;
		InitOptionPanel initOptionPanel = new InitOptionPanel(setting, initOnly);
		updateSettings.add(initOptionPanel);
		if(initOnly){
			main = initOptionPanel;
		}else{
			AnalysisOptionPanel analysisOptionPanel = new AnalysisOptionPanel(setting);
			updateSettings.add(analysisOptionPanel);
			
			//タブわけ
			JTabbedPane tabbedPane = new JTabbedPane();
			tabbedPane.add(analysisOptionPanel, Messages.getString("Setting.sect.general")); //$NON-NLS-1$
			tabbedPane.add(initOptionPanel, Messages.getString("Setting.sect.core")); //$NON-NLS-1$
			
			main = tabbedPane;
		}
		
		
		Box mainBox = Box.createVerticalBox();
		mainBox.add(main);
		mainBox.add(Box.createVerticalStrut(10));
		mainBox.add(buttonPanel);
		mainBox.setBorder(new EmptyBorder(10,10,10,10));
		
		dialog.getContentPane().add(mainBox);
		dialog.pack();
		dialog.setLocationRelativeTo(parent);
	}
	
	/**
	 * 設定を渡されたsettingに登録します。
	 * @param setting
	 * @return false if failed
	 */
	private boolean update(Setting setting){
		try {
			for(UpdateSetting us : updateSettings)
				us.update(setting);
		} catch (InvalidOptionException e) {
			JOptionPane.showMessageDialog(dialog.getParent(), e.getMessage(), Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$
			return false;
		}
		return true;
	}
	
	public Setting getSetting() {
		return setting;
	}

	private static class InvalidOptionException extends Exception{
		InvalidOptionException(String message){
			super(message);
		}
	}
	
	private interface UpdateSetting{
		void update(Setting setting) throws InvalidOptionException;
	}
	
	private class UpdateSettingContainer extends JPanel implements UpdateSetting{

		private List<UpdateSetting> updateSettings = new ArrayList<UpdateSetting>();
		
		@Override
		public Component add(Component comp) {
			if (comp instanceof UpdateSetting) {
				UpdateSetting us = (UpdateSetting) comp;
				addUpdateSetting(us);
			}
			return super.add(comp);
		}
		
		protected void addUpdateSetting(UpdateSetting updateSetting){
			updateSettings.add(updateSetting);
		}
		
		public void update(Setting setting) throws InvalidOptionException {
			for(UpdateSetting us : updateSettings)
				us.update(setting);
		}
	}
	
	private class InitOptionPanel extends UpdateSettingContainer{

		InitOptionPanel(Setting setting, boolean initOnly){
			setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
			setBorder(new EmptyBorder(10,10,10,10));
			
			BaseOptionPanel baseOptionPanel = new BaseOptionPanel(setting);
			NewOptionPanel newOptionPanel = new NewOptionPanel(setting);
			UpdateOptionPanel updateOptionPanel = new UpdateOptionPanel(setting);
			LanguageOptionPanel languageOptionPanel = new LanguageOptionPanel(setting);
			
			JLabel notice = new JLabel(Messages.getString("Setting.valid_after_restart")); //$NON-NLS-1$
			Box noticeBox = Box.createHorizontalBox();
			noticeBox.add(Box.createHorizontalGlue());
			noticeBox.add(notice);
			
			add(baseOptionPanel);
			//add(Box.createVerticalStrut(5));
			add(newOptionPanel);
			//add(Box.createVerticalStrut(5));
			add(updateOptionPanel);
			//add(Box.createVerticalStrut(5));
			add(languageOptionPanel);
			if(!initOnly){
				add(Box.createVerticalStrut(10));
				add(noticeBox);
			}
		}
		
		private class BaseOptionPanel extends JPanel implements UpdateSetting{
		
			private JTextField dirField;
			private JCheckBox useLibraryPack;
			
			BaseOptionPanel(Setting setting){
				setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
				setBorder(new TitledBorder(Messages.getString("Setting.r_prop"))); //$NON-NLS-1$
				
				//Ｒのディレクトリ
				dirField = new JTextField((String)setting.getProperty(Setting.Key.R_HOME));
				dirField.setMaximumSize(new Dimension(Integer.MAX_VALUE, 20));
				dirField.setPreferredSize(new Dimension(400, 20));
				JButton ref = new JButton(createDirSelectionAction(this, dirField));
		
				Box dirBox = Box.createHorizontalBox();
				//dirBox.setBorder(new TitledBorder("Ｒのディレクトリ"));
				dirBox.add(new JLabel(Messages.getString("Setting.r_home"))); //$NON-NLS-1$
				dirBox.add(dirField);
				dirBox.add(ref);
				
				//ライブラリパックの使用
				useLibraryPack = new JCheckBox(Messages.getString("Setting.use_libpack"), (Boolean)setting.getProperty(Setting.Key.USE_LIBLARY_PACK)); //$NON-NLS-1$
				Box useLPBox = Box.createHorizontalBox();
				useLPBox.add(useLibraryPack);
				useLPBox.add(Box.createHorizontalGlue());
				
				
				add(dirBox);
				add(Box.createVerticalStrut(5));
				add(useLPBox);
			}
			
			public void update(Setting setting) throws InvalidOptionException {
				File dir = new File(dirField.getText());
				//TODO R.dllなどの確認？
				if(!dir.exists() || !dir.isDirectory())
					throw new InvalidOptionException(Messages.getString("Error.Setting.r_home")); //$NON-NLS-1$
				
				setting.setProperty(Setting.Key.R_HOME, dir.getAbsolutePath());
				setting.setProperty(Setting.Key.USE_LIBLARY_PACK, useLibraryPack.isSelected());
			}
		}

		private class UpdateOptionPanel extends JPanel implements UpdateSetting{
			
			private JCheckBox checkUpdate;
			
			UpdateOptionPanel(Setting setting){
				super();
				setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
				//setBorder(new EmptyBorder(0,10,0,10));
				setBorder(new TitledBorder(Messages.getString("Setting.update"))); //$NON-NLS-1$
				
				checkUpdate = new JCheckBox(Messages.getString("Setting.check_update") //$NON-NLS-1$
						, (Boolean)setting.getProperty(Setting.Key.IS_AUTO_UPDATE));
				Box checkUpdateBox = Box.createHorizontalBox();
				checkUpdateBox.add(checkUpdate);
				checkUpdateBox.add(Box.createHorizontalGlue());
				
				add(checkUpdateBox);
			}
		
			public void update(Setting setting) throws InvalidOptionException{
				setting.setProperty(Setting.Key.IS_AUTO_UPDATE, checkUpdate.isSelected());
			}
		}

		private class LanguageOptionPanel extends JPanel implements UpdateSetting{
			private JComboBox language;
			
			public LanguageOptionPanel(Setting setting) {
				super();
				setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
				//setBorder(new EmptyBorder(0,10,0,10));
				setBorder(new TitledBorder(Messages.getString("Setting.language"))); //$NON-NLS-1$
				
				//JLabel label = new JLabel(Messages.getString("Setting.language"));
				
				Locale[] langs = Messages.getLocales();
				//Locale[] langs = new Locale[_langs.length+1];
				//langs[0] = Locale.ENGLISH;
				//for(int i = 1; i < langs.length; i++)
				//	langs[i] = _langs[i-1];
				language = new JComboBox(langs);
				language.setRenderer(new DefaultListCellRenderer(){
					@Override
					public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
						super.getListCellRendererComponent(list, value, index, isSelected,
								cellHasFocus);
						
						Locale locale = (Locale)value;
						setText(locale.getDisplayLanguage());
						
						return this;
					}
				});
				language.setSelectedItem(Locale.ENGLISH);
				language.setSelectedItem(new Locale((String)setting.getProperty(Setting.Key.LANGUAGE)));
				language.setPreferredSize(new Dimension(100, 20));
				language.setMaximumSize(new Dimension(100, 20));
				
				//JLabel langLabel = new JLabel(Messages.getString("Setting.valid_after_restart")); //$NON-NLS-1$
				
				Box langBox = Box.createHorizontalBox();
				//langBox.add(label);
				//langBox.add(Box.createHorizontalStrut(10));
				langBox.add(language);
				//langBox.add(langLabel);
				langBox.add(Box.createHorizontalGlue());
				
				add(langBox);
			}
			
			public void update(Setting setting) throws InvalidOptionException {
				Locale locale = (Locale)language.getSelectedItem();
				setting.setProperty(Setting.Key.LANGUAGE, locale.getLanguage());
			}
		}

		private class NewOptionPanel extends JPanel implements UpdateSetting{
			
			private JTextField dirField;
			
			NewOptionPanel(Setting setting){
				super();
				setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
				//setBorder(new EmptyBorder(0,10,0,10));
				setBorder(new TitledBorder(Messages.getString("Setting.working_directory"))); //$NON-NLS-1$
				
				/*//
				JLabel dirLabel = new JLabel("起動時の作業ディレクトリ");
				Box labelBox = Box.createHorizontalBox();
				labelBox.add(dirLabel);
				labelBox.add(Box.createHorizontalGlue());
				*/
				
				//デフォルトの作業ディレクトリ
				dirField = new JTextField((String)setting.getProperty(Setting.Key.DEFAULT_WORKING_DIRECTORY));
				dirField.setMaximumSize(new Dimension(Integer.MAX_VALUE, 20));
				dirField.setPreferredSize(new Dimension(400, 20));
				JButton ref = new JButton(createDirSelectionAction(this, dirField));
				
				Box dirBox = Box.createHorizontalBox();
				//dirBox.setBorder(new TitledBorder(Messages.getString("Setting.working_directory"))); //$NON-NLS-1$
				dirBox.add(dirField);
				dirBox.add(ref);
				
				
				//add(labelBox);
				add(dirBox);
			}
		
			public void update(Setting setting) throws InvalidOptionException{
				if(dirField.getText().equals("")){ //$NON-NLS-1$
					setting.setProperty(Setting.Key.DEFAULT_WORKING_DIRECTORY, ""); //$NON-NLS-1$
				}else{
					File dir = new File(dirField.getText());
					if(!dir.exists() || !dir.isDirectory())
						throw new InvalidOptionException(Messages.getString("Error.invalid_workingdir")); //$NON-NLS-1$
					setting.setProperty(Setting.Key.DEFAULT_WORKING_DIRECTORY, dir.getAbsolutePath());
				}
			}
		}
	}
	

	private class AnalysisOptionPanel extends UpdateSettingContainer{

		AnalysisOptionPanel(Setting setting){
			setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
			setBorder(new EmptyBorder(10,10,10,10));
			
			NewOptionPanel newOptionPanel = new NewOptionPanel(setting);
			ConsoleOptionPanel consoleOptionPanel = new ConsoleOptionPanel(setting);

			addUpdateSetting(consoleOptionPanel);
			addUpdateSetting(newOptionPanel);

			add(consoleOptionPanel);
			add(newOptionPanel);
		}

		private class NewOptionPanel extends JPanel implements UpdateSetting{
			
			private JCheckBox useGrid;
			private JSpinner gridDistance;
			
			NewOptionPanel(Setting setting){
				super();
				setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
				setBorder(new TitledBorder(Messages.getString("Setting.new_flow_defaults"))); //$NON-NLS-1$
				
				//グリッドのon/off
				useGrid = new JCheckBox(Messages.getString("Setting.use_grid"),  //$NON-NLS-1$
						(Boolean)setting.getProperty(Setting.Key.USE_GRID));
				Box useGridBox = Box.createHorizontalBox();
				useGridBox.add(useGrid);
				useGridBox.add(Box.createHorizontalGlue());
				//グリッドの幅
				JLabel gridDistLabel = new JLabel(Messages.getString("Setting.grid_distance"));  //$NON-NLS-1$
				int value = (Integer)setting.getProperty(Setting.Key.GRID_DISTANCE);
				gridDistance = new JSpinner(new SpinnerNumberModel(value, 0, 500, 1));
				gridDistance.setMaximumSize(new Dimension(50, Integer.MAX_VALUE));
				gridDistance.setPreferredSize(new Dimension(100, 20));
				gridDistance.setMaximumSize(gridDistance.getPreferredSize());
				Box gridDistBox = Box.createHorizontalBox();
				gridDistBox.setMaximumSize(new Dimension(Integer.MAX_VALUE, 20));
				gridDistBox.add(gridDistLabel);
				gridDistBox.add(Box.createHorizontalStrut(10));
				gridDistBox.add(gridDistance);
				gridDistBox.add(Box.createHorizontalGlue());
		
				Box gridBox = Box.createVerticalBox();
				gridBox.setBorder(new TitledBorder(Messages.getString("Setting.grid"))); //$NON-NLS-1$
				gridBox.add(useGridBox);
				gridBox.add(Box.createVerticalStrut(10));
				gridBox.add(gridDistBox);
				
				add(gridBox);
			}
		
			public void update(Setting setting) throws InvalidOptionException{		
				setting.setProperty(Setting.Key.USE_GRID, useGrid.isSelected());
				setting.setProperty(Setting.Key.GRID_DISTANCE, (Integer)gridDistance.getValue());
			}
		}

		private class ConsoleOptionPanel extends JPanel implements UpdateSetting{
		
			private JCheckBox useBuffer;
			
			ConsoleOptionPanel(Setting setting){
				super();
				setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
				setBorder(new TitledBorder(Messages.getString("Setting.console"))); //$NON-NLS-1$
				
				useBuffer = new JCheckBox(Messages.getString("Setting.use_buffer"), (Boolean)setting.getProperty(Setting.Key.USE_CONSOLE_BUFFER)); //$NON-NLS-1$
				Box checkUpdateBox = Box.createHorizontalBox();
				checkUpdateBox.add(useBuffer);
				checkUpdateBox.add(Box.createHorizontalGlue());
				
				add(checkUpdateBox);
			}
		
			public void update(Setting setting) throws InvalidOptionException{
				setting.setProperty(Setting.Key.USE_CONSOLE_BUFFER, useBuffer.isSelected());
			}
		}
	}
	
	private static Action createDirSelectionAction(final Component parent, final JTextComponent textComp){
		return new AbstractAction("..."){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				File file = FileChooserUtility.chooseDir(parent, 
						new File(textComp.getText()));
				if(file == null)return;
				textComp.setText(file.getAbsolutePath());
			}
		};
	}
	
	public static Setting showSettingDialog(Window parent, Setting orginal, boolean initOnly){
		SettingDialog sd = new SettingDialog(parent, orginal, initOnly);
		sd.dialog.setVisible(true);
		sd.dialog.dispose();
		return sd.getSetting();
	}
	
	public static void main(String[] args) {
		//setLookAndFeel
		new JPanel();
        try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e) {}
		
		//設定ファイル準備
		FileResource.prepareFiles();
		Setting setting = new Setting();

		//設定ファイルの読み込み（設定ファイルが存在する場合）
		if(FileResource.SETTING_FILE.exists()){
			try {
				PropertiesUtility.load(setting, FileResource.SETTING_FILE);
				setting.update();
			} catch (IOException e2) {
				//e2.printStackTrace();
				JOptionPane.showMessageDialog(null, Messages.getString("Error.read_option"), Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE);  //$NON-NLS-1$//$NON-NLS-2$
			}
		}
		
		//ロケールの設定
		Locale.setDefault(new Locale(((String)setting.getProperty(Key.LANGUAGE))));
		
		//設定ダイアログ
		JFrame dummy = new JFrame(Messages.getString("Action.preferences")); //$NON-NLS-1$
		dummy.setIconImage(FileResource.DESKTOP_16.getImage());
		dummy.setUndecorated(true);
		dummy.setLocationRelativeTo(null);
		dummy.setVisible(true);
		Setting newSetting = SettingDialog.showSettingDialog(dummy, setting, true);
		dummy.dispose();
		
		//設定の保存
		if(newSetting != null){
			try {
				PropertiesUtility.store(newSetting, FileResource.SETTING_FILE);
			} catch (Exception e1) {
				JOptionPane.showMessageDialog(null, Messages.getString("Error.write_option"), Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$ //$NON-NLS-2$
				return;
			}
		}
	}
}
