package com.ef_prime.rflow.console;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FileDialog;
import java.awt.Frame;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.KeyStroke;
import javax.swing.border.EmptyBorder;
import javax.swing.border.LineBorder;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;

import org.rosuda.JRI.RMainLoopCallbacks;
import org.rosuda.JRI.Rengine;

import com.ef_prime.rflow.CommandReceiver;
import com.ef_prime.rflow.Constants;
import com.ef_prime.rflow.FileResource;
import com.ef_prime.rflow.Prefs;
import com.ef_prime.rflow.RCommand;
import com.ef_prime.rflow.Rflow;
import com.ef_prime.rflow.Setting;
import com.ef_prime.rflow.ActionManager.ActionKey;
import com.ef_prime.rflow.exception.IncompatibleCommandException;
import com.ef_prime.rflow.flow.model.FlowNode;
import com.ef_prime.rflow.node.base.SentenceNodeModel;
import com.ef_prime.rflow.resource.Messages;
import com.ef_prime.rflow.tool.CodingTextPane;
import com.ef_prime.rflow.util.NoWrapEditorKit;
import com.ef_prime.rflow.util.SizeLimitVector;
import com.ef_prime.rflow.util.SwingUtils;

public class RConsole implements RMainLoopCallbacks, PropertyChangeListener{

	/**
	 * コマンドの履歴
	 */
	private final SizeLimitVector<String> commandHistory = new SizeLimitVector<String>(512);

	private int histPos = commandHistory.size();
	
	//private boolean isBusy;
	
	private Rflow rflow; 
	
//	private JDialog dialog;
	private JFrame dialog;
	private JTextPane output;
	private CodingTextPane input;
	private JScrollPane outputSp;
	private CommandReceiver receiver;
	
	private boolean useBuffer;
	private static final int BUFFER_SIZE = 1024;//number of char
	private StringBuffer buffer = new StringBuffer(BUFFER_SIZE);
	private int currentOutputType = 0;
	
	public RConsole(Rflow rflow){
		this.rflow = rflow;
		this.receiver = rflow.receiver;
		useBuffer = (Boolean)Rflow.setting.getProperty(Setting.Key.USE_CONSOLE_BUFFER);
		Rflow.setting.addPropertyChangeListener(Setting.Key.USE_CONSOLE_BUFFER, this);
		
		output = new JTextPane();
		output.setFont(Prefs.CODE_FONT);
		output.setEditable(false);
		output.setEditorKit(new NoWrapEditorKit());
		SwingUtils.setTabs(output, Prefs.CHAR_PER_TAB);
		
		outputSp = new JScrollPane(output, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		outputSp.setBorder(new EmptyBorder(0,0,0,0));
		
		input = new CodingTextPane();
		input.setIntend(false);
		input.setPreferredSize(new Dimension(400, 50));
		input.setMaximumSize(new Dimension(400, 50));
		input.addKeyListener(new InputKeyHandler());
		input.setBorder(new LineBorder(Color.GRAY, 1));
		
		//dialog = new JDialog(rflow.frame, "R console", ModalityType.MODELESS){
//		dialog = new JDialog(rflow.frame, "R console"){
		dialog = new JFrame("R console"); //$NON-NLS-1$
//		{
//			@Override
//			public void setVisible(boolean b) {
//				super.setVisible(b);
//				Actions.CONSOLE_WINDOW.putValue(Actions.SELECTED_KEY, b);
//			}
//		};
		dialog.setIconImage(FileResource.DESKTOP_16.getImage());
		dialog.getContentPane().setLayout(new BorderLayout());
		dialog.getContentPane().add(createMenuBar(), BorderLayout.NORTH);
		dialog.getContentPane().add(outputSp, BorderLayout.CENTER);
		dialog.getContentPane().add(input, BorderLayout.SOUTH);
		dialog.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		//dialog.setDefaultCloseOperation(JFrame.HIDE_ON_CLOSE);
		//dialog.setModalExclusionType(ModalExclusionType.APPLICATION_EXCLUDE);
		//サイズの変更をＲに通知
		dialog.addComponentListener(new ComponentAdapter(){
			@Override
			public void componentResized(ComponentEvent e) {
				int charWidth = output.getFontMetrics(output.getFont()).charWidth('R');
				
				int width = output.getVisibleRect().width / charWidth - 1;
				if(width < 10)width = 10;
				else if(width > 10000)width = 10000;
				if(RCommand.getR() != null)
					RCommand.getR().idleEval("options(width="+width+")"); //$NON-NLS-1$ //$NON-NLS-2$
			}
		});
		
		dialog.setSize(600, 400);
		dialog.setLocation(0, 0+400);
		//dialog.setLocationRelativeTo(parent);
		//dialog.setVisible(true);
	}
	
	private JMenuBar createMenuBar(){
		final int mask = Toolkit.getDefaultToolkit().getMenuShortcutKeyMask();
		
		JMenuBar menuBar = new JMenuBar();
		
		JMenu command = new JMenu(Messages.getString("Menu.command")); //$NON-NLS-1$
		
		Action submit = new AbstractAction(Messages.getString("Action.run")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				submitCommand();
			}
		};
		submit.putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0));
		
		Action create = new AbstractAction(Messages.getString("Action.new_node")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				createNewNode();
			}
		};
		create.putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, mask));
		
		command.add(submit);
		command.add(create);
		command.add(rflow.actionManager.getAction(ActionKey.STOP));
		
//		JMenu display = new JMenu("表示");
//		display.add(Actions.CONSOLE_WINDOW);

		menuBar.add(command);
//		menuBar.add(display);
		
		return menuBar;
	}
	
	public boolean isVisible(){
//		return dialog.getHeight() > 50;
		return dialog.isVisible();
	}
	
	//private int previousHeight = 400;
	public void setVisible(boolean visible){
		//if(visible)
		dialog.setVisible(visible);
		/*
		if(visible)
			dialog.setSize(dialog.getWidth(), previousHeight);
		else
			dialog.setSize(dialog.getWidth(), 0);
			*/
	}
	
	public void clearConsole(){
		try {
			Document outputDoc = output.getDocument();
			int len = outputDoc.getLength();
			if(len > 1 && outputDoc.getText(len-2, 2).equals("> ")) //$NON-NLS-1$
				len -= 2;
			outputDoc.remove(0, len);
			input.getDocument().remove(0, input.getDocument().getLength());
		} catch (BadLocationException e) {
			e.printStackTrace();
		}
	}
	
	private void append(String text, AttributeSet attributeSet){
    	Document document = output.getDocument();
    	try {
			document.insertString(document.getLength(), text, attributeSet);
		} catch (BadLocationException e) {
			e.printStackTrace();
		}		
		
		if(document.getLength() > Prefs.CONSOLE_CHAR_MAX)
		{
			int delChar = document.getLength() - Prefs.CONSOLE_CHAR_MAX;
			try {
				String s = document.getText(0, document.getLength());
				
				int index = s.indexOf(Constants.NEWLINE, delChar) + 1;
				if(index > 0){
					document.remove(0, index);
				}else{
					//改行なしで制限超えた場合
					document.remove(0, delChar);
				}
			} catch (BadLocationException e) {
				e.printStackTrace();
			}
		}
		
		output.setCaretPosition(document.getLength());
	}
	
	private void addToHistory(String command){
		String s = command.trim();
		if(s.length() > 0){
	    	commandHistory.add(command.trim());
	    	histPos = commandHistory.size();
		}
	}
	
	private void submitCommand(){
        String command = input.getRawText().trim();
        input.setText(""); //$NON-NLS-1$
    	receiver.passCommand(command);
	}
	
	private void createNewNode(){
        String command = input.getText().trim();
        input.setText(""); //$NON-NLS-1$
    	//複数行を許す場合
    	/*NodeModel[] models = RCommand.parseToNodes(command);
    	if(models != null){
    		ViewLayout layout = new ViewLayout(models);
    		rflow.viewer.addSubLayout(layout, new Point(0,0));
    	}*/
    	SentenceNodeModel model = new SentenceNodeModel();
    	try {
    		if(command.length() < 1 && histPos == commandHistory.size())
    			command = commandHistory.get(histPos - 1);
			model.setCommand(command);
			rflow.viewer.addNode(new FlowNode(model), null, true);
		} catch (IncompatibleCommandException e1) {
			// TODO 再読み込み検討
			e1.printStackTrace();
		}
    	
    	//histPos = commandHistory.size();
		addToHistory(command);
	}
	
	private void flushBuffer(){
    	if(buffer.length() > 0){
	    	if(currentOutputType > 0)
	    		append(buffer.toString(), Prefs.ERROR_ATTR);
	    	else
	    		append(buffer.toString(), Prefs.OUTPUT_ATTR);
	    	buffer.delete(0, buffer.length());
    	}
	}
    
    public void rBusy(Rengine re, int which) {
    	if(Rflow.DEBUG > 0)
    		System.out.println("rBusy("+which+")"); //$NON-NLS-1$ //$NON-NLS-2$
        //isBusy = (which == 1);
    }
    
    public void rWriteConsole(Rengine re, String text, int oType) {
    	if(useBuffer){
	    	if(currentOutputType != oType)
	            flushBuffer();
	    	currentOutputType = oType;
	    	
	    	if(buffer.length()+text.length() > buffer.capacity())
	            flushBuffer();
	    		
	    	buffer.append(text);
    	}else{
	    	if(oType > 0)
	    		append(text, Prefs.ERROR_ATTR);
	    	else
	    		append(text, Prefs.OUTPUT_ATTR);
    	}
    }
    
    public String rReadConsole(Rengine re, String prompt, int addToHistory) {
        if(useBuffer)flushBuffer();
    	append(prompt, Prefs.INPUT_ATTR);
    	//複数plotのreturnおくり
    	/*if(isBusy){
    		waitReturn();
    		rWriteConsole(re, Constants.NEWLINE, 0);
    		return Constants.NEWLINE;
    	}*/
        try {
        	String s = receiver.waitForCommand();
        	append(s, Prefs.INPUT_ATTR);
        	addToHistory(s);
        	
            return (s==null||s.length()==0)?s:s+Constants.NEWLINE_STR;
        } catch (Exception e) {
            System.out.println("jriReadConsole exception: "+e.getMessage()); //$NON-NLS-1$
        }
        return null;
    }
    
    public void rShowMessage(Rengine re, String message) {
    	if(Rflow.DEBUG > 0)
    		System.out.println("rShowMessage \""+message+"\""); //$NON-NLS-1$ //$NON-NLS-2$
    }
	
    public String rChooseFile(Rengine re, int newFile) {
		FileDialog fd = new FileDialog(new Frame(), (newFile==0)?"Select a file":"Select a new file", (newFile==0)?FileDialog.LOAD:FileDialog.SAVE);
		fd.setVisible(true);
		String res=null;
		if (fd.getDirectory()!=null) res=fd.getDirectory();
		if (fd.getFile()!=null) res=(res==null)?fd.getFile():(res+fd.getFile());
		return res;
    }
    
    public void   rFlushConsole (Rengine re) {
    	flushBuffer();
    }
	
    public void   rLoadHistory  (Rengine re, String filename) {
    }			
    
    public void   rSaveHistory  (Rengine re, String filename) {
    	//System.out.println("save history");
    }

    /*--------------------------------------------------------------*/
    //wait return
    /*
    private boolean hitReturn = false;
    private boolean waitingReturn = false;
    synchronized void waitReturn(){
		//System.out.println("wait for return!");
		waitingReturn = true;
    	while(!hitReturn){
			try {
				wait(200);
			} catch (InterruptedException e) {
			}
    	}
    	hitReturn = false;
    	waitingReturn = false;
    }
    
    synchronized void notifyReturn(){
    	hitReturn = true;
    	notifyAll();
    }
    */
    /*--------------------------------------------------------------*/

    private class InputKeyHandler extends KeyAdapter{
    	@Override
    	public void keyPressed(KeyEvent e) {
    		if(e.isConsumed())return;
    		
    		CodingTextPane console = (CodingTextPane)e.getSource();
    		switch(e.getKeyCode()){
    		case KeyEvent.VK_ENTER:
//    			if(waitingReturn){
//    				notifyReturn();
//    				e.consume();
//    				return;
//    			}
                if(e.isControlDown()){
                	createNewNode();
                }else{
                	submitCommand();                	
                }
                e.consume();
    			break;
    		case KeyEvent.VK_UP:
    			if(--histPos < 0){
    				histPos = 0;
    				console.setText(""); //$NON-NLS-1$
    				return;
    			}
    			console.setText(getHistory(histPos));
    			e.consume();
    			break;
    		case KeyEvent.VK_DOWN:
    			if(++histPos > commandHistory.size()){
    				histPos = commandHistory.size();
    				return;
    			}
    			console.setText(getHistory(histPos));
    			e.consume();
    			break;
    		case KeyEvent.VK_ESCAPE:
    			//中断
    			//receiver.passCommand("\u001B");
    			RCommand.getR().rniStop(0);
    			receiver.passCommand(""); //$NON-NLS-1$
    			e.consume();
    			break;
    		}
    	}
    	
    	private String getHistory(int histPos){
    		if(histPos == commandHistory.size())
    			return ""; //$NON-NLS-1$
			return commandHistory.get(histPos);
    	}
    }

    /*-------------PropertyChangeListener-------------*/
	public void propertyChange(PropertyChangeEvent e) {
		if(Setting.Key.USE_CONSOLE_BUFFER.getKeyString().equals(e.getPropertyName())){
			Setting setting = (Setting)e.getSource();
			flushBuffer();
			useBuffer = (Boolean)setting.getProperty(Setting.Key.USE_CONSOLE_BUFFER);
		}
	}
}
