package com.ef_prime.rflow.flow;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetAdapter;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DropTargetEvent;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.image.ImageObserver;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPopupMenu;
import javax.swing.SwingUtilities;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.undo.CompoundEdit;
import javax.swing.undo.UndoManager;

import com.ef_prime.rflow.Constants;
import com.ef_prime.rflow.FlowSetting;
import com.ef_prime.rflow.Prefs;
import com.ef_prime.rflow.RCommand;
import com.ef_prime.rflow.Rflow;
import com.ef_prime.rflow.RflowUtil;
import com.ef_prime.rflow.ActionManager.ActionKey;
import com.ef_prime.rflow.exception.CanceledException;
import com.ef_prime.rflow.exception.IncompatibleCommandException;
import com.ef_prime.rflow.explorer.FunctionNode;
import com.ef_prime.rflow.explorer.FunctionNodeFlavor;
import com.ef_prime.rflow.flow.model.FlowEdge;
import com.ef_prime.rflow.flow.model.FlowGraph;
import com.ef_prime.rflow.flow.model.FlowLayout;
import com.ef_prime.rflow.flow.model.FlowNode;
import com.ef_prime.rflow.flow.rendering.FlowRenderer;
import com.ef_prime.rflow.node.NodeModel;
import com.ef_prime.rflow.node.NodeModelChangeListener;
import com.ef_prime.rflow.node.NodeModelController;
import com.ef_prime.rflow.node.NodeModelFactory;
import com.ef_prime.rflow.node.NodeModelManager;
import com.ef_prime.rflow.node.base.FreeNodeModel;
import com.ef_prime.rflow.node.base.SentenceNodeModel;
import com.ef_prime.rflow.node.base.SubflowNodeModel;
import com.ef_prime.rflow.resource.Messages;
import com.ef_prime.rflow.util.AbstractActionManager;
import com.ef_prime.rflow.util.NamedCompoundEdit;

import edu.uci.ics.jung.graph.ArchetypeVertex;
import edu.uci.ics.jung.graph.Edge;
import edu.uci.ics.jung.graph.Vertex;
import edu.uci.ics.jung.graph.decorators.ToolTipFunction;
import edu.uci.ics.jung.graph.event.GraphEvent;
import edu.uci.ics.jung.graph.event.GraphEventListener;
import edu.uci.ics.jung.graph.event.GraphEventType;
import edu.uci.ics.jung.visualization.Layout;
import edu.uci.ics.jung.visualization.ShapePickSupport;
import edu.uci.ics.jung.visualization.VisualizationModel;
import edu.uci.ics.jung.visualization.VisualizationViewer;
import edu.uci.ics.jung.visualization.control.PluggableGraphMouse;

//TODO undo動作をlayoutにうつす
public class FlowViewer extends VisualizationViewer implements
		NodeModelChangeListener {

	private Rflow rflow;

	private FlowExecuter executer;

	private static ViewLayout clipedLayout;

	private UndoManager undoManager;

	UndoableEditFactory undoableEditFactory;

	private boolean useGrid;

	private List<UndoableEditListener> undoableEditListeners = new ArrayList<UndoableEditListener>();

	public FlowViewer(Rflow rflow, FlowState state) {
		super(new FlowLayout(new FlowGraph()), new FlowRenderer(state));
		this.rflow = rflow;
		executer = new FlowExecuter(rflow.receiver, state);
		undoManager = new FlowUndoManager(this);
		undoableEditFactory = new UndoableEditFactory(this);

		addUndoableEditListener(undoManager);

		setBackground(Color.WHITE);
		setPickSupport(new ShapePickSupport(10));
		setPickedState(state);

		setGraphEventListeners();
		// フラグノードが変わったらローカル環境をクリアする
		state.addFlagNodeChageListener(new FlowState.FlagNodeChangeListener() {
			public void flagNodeChanged(FlowNode newFlagNode) {
				RCommand.clearLocalEnv();
			}
		});

		// マウス
		/*
		 * AdjustedPluggableGraphMouse do not invoke mouseClicked mouseEntered
		 * mouseExited mouseMoved
		 */
		PluggableGraphMouse graphMouse = new AdjustedPluggableGraphMouse();
		graphMouse.add(new FlowPickingGraphMousePlugin(MouseEvent.BUTTON1_MASK,
				MouseEvent.BUTTON1_MASK | MouseEvent.CTRL_MASK));
		/*graphMouse.add(new FlowTranslatingGraphMousePlugin(
				MouseEvent.BUTTON1_MASK | MouseEvent.SHIFT_MASK));
		graphMouse.add(new ScalingGraphMousePlugin(new LayoutScalingControl(),
				MouseEvent.NOBUTTON));
		graphMouse.add(new ScalingGraphMousePlugin(new ViewScalingControl(),
				MouseEvent.NOBUTTON));
		graphMouse.add(new RotatingGraphMousePlugin(MouseEvent.BUTTON1_MASK
				| MouseEvent.ALT_MASK));*/
		setGraphMouse(graphMouse);

		addMouseListener(new MouseHandler());

		// ツールチップ
		setToolTipFunction(new ToolTipHandler());

		// ドロップ動作
		setDropTarget(new DropTarget(this, new DropTargetHandler()));

		updateFileSetting();
	}
	
	public FlowViewer createNewViewer() {
		return new FlowViewer(rflow, getPickedState());
	}

	public void clearFlow() {
		setGraphLayout(new FlowLayout(new FlowGraph()));
		getPickedState().clearAllState();
		undoManager.discardAllEdits();
		updateUndoState();
		repaint();
	}

	public void init(ViewLayout viewLayout) {
		FlowLayout layout = new FlowLayout(viewLayout.getGraph());
		setGraphLayout(layout);
		layoutTransformer.setToIdentity();
		for (Map.Entry<Vertex, Point2D> entry : viewLayout.getLocationMap().entrySet()) {
			Point2D p = entry.getValue();
			layout.forceMove(entry.getKey(), p.getX(), p.getY());
		}

		Point offset = viewLayout.getOffset();
		if (offset != null)
			layoutTransformer.translate(offset.x, offset.y);

		updateFileSetting();
	}

	public void updateFileSetting() {
		FlowSetting fileSettings = rflow.getFileSetting();
		FlowLayout layout = getGraphLayout();
		layout.setGridDistance((Integer) fileSettings
				.getProperty(FlowSetting.Key.GRID_DISTANCE));

		setUseGrid((Boolean) fileSettings.getProperty(FlowSetting.Key.USE_GRID));
	}

	public void nodeModelChanged(FlowNode node, NodeModel old, NodeModel model) {
		fireUndoableEditUpdate(new UndoableEditEvent(this, undoableEditFactory
				.createNodeModelChangeEdit(node, old)));
		updateGroup(model);
		repaint();
	}

	@Override
	public void setGraphLayout(Layout layout) {
		setGraphLayout(layout, false);
	}

	@Override
	public void setGraphLayout(Layout layout, boolean scaleToLayout) {
		super.setGraphLayout(layout, scaleToLayout);
		setGraphEventListeners();
		updateGroup();
	}

	@Override
	public void setModel(VisualizationModel model) {
		super.setModel(model);
		setGraphEventListeners();
	}

	private void setGraphEventListeners() {
		// TODO
		getGraph().addGraphWillChangeListener(getPickedState());
		getGraph().addListener(new GraphEventHandler(),
				GraphEventType.ALL_SINGLE_EVENTS);
		if (rflow.codeViewer != null) {
			getGraph().addListener(rflow.codeViewer,
					GraphEventType.ALL_SINGLE_EVENTS);
			rflow.codeViewer.flowModelChanged();
		}
	}

	public FlowGraph getGraph() {
		return (FlowGraph) getGraphLayout().getGraph();
	}

	@Override
	public FlowRenderer getRenderer() {
		return (FlowRenderer) super.getRenderer();
	}

	@Override
	public FlowLayout getGraphLayout() {
		return (FlowLayout) super.getGraphLayout();
	}

	@Override
	public FlowState getPickedState() {
		return (FlowState) super.getPickedState();
	}

	public void setUseGrid(boolean useGrid) {
		this.useGrid = useGrid;
	}

	public boolean isUseGrid() {
		return useGrid;
	}

	public void updateGroup() {
		for (Object o : getGraph().getVertices()) {
			updateGroup(((FlowNode) o).getModel());
		}
		repaint();
	}

	public void updateGroup(NodeModel model) {
		model.update(rflow.groupManager);
	}

	public void setDirty(boolean isDirty) {
		rflow.setDirty(isDirty);
	}

	public List<UndoableEditListener> getUndoableEditListeners() {
		return undoableEditListeners;
	}

	public void addUndoableEditListener(UndoableEditListener listener) {
		undoableEditListeners.add(listener);
	}

	public void removeUndoableEditListener(UndoableEditListener listener) {
		undoableEditListeners.remove(listener);
	}

	void fireUndoableEditUpdate(UndoableEditEvent e) {
		for (UndoableEditListener l : undoableEditListeners) {
			l.undoableEditHappened(e);
		}
	}

	/**
	 * nodeまでのフローを実行します。
	 * 
	 * @param node
	 * @param clear
	 */
	public void execute(FlowNode node, boolean clear) {
		if (node == null)
			return;
		executer.execute(node, clear);
		getPickedState().setFlagNode(node);
		repaint();
	}

	/**
	 * フラグノードの次からnodeまでを実行
	 * 
	 * @param node
	 */
	public void executeRest(FlowNode node) {
		if (node == null)
			return;

		List<FlowNode> stream = FlowUtil.getStream(node);
		if (stream.contains(getPickedState().getFlagNode())) {
			executer.excuteRest(node);
			getPickedState().setFlagNode(node);
			repaint();
		} else
			execute(node, false);
	}

	/**
	 * nodeまでのフローのスクリプトをコピーします。
	 * 
	 * @param node
	 */
	public void copyScript(FlowNode node) {
		if (node == null)
			return;
		executer.copyCommandToClipBoard(node);
	}

	public void undo() {
		if (undoManager.canUndo())
			undoManager.undo();
		updateUndoState();
	}

	public void redo() {
		if (undoManager.canRedo())
			undoManager.redo();
		updateUndoState();
	}

	void updateUndoState() {
		rflow.actionManager.getAction(ActionKey.UNDO).setEnabled(
				undoManager.canUndo());
		rflow.actionManager.getAction(ActionKey.REDO).setEnabled(
				undoManager.canRedo());

		rflow.actionManager.getAction(ActionKey.UNDO).putValue(Action.NAME,
				undoManager.getUndoPresentationName());
		rflow.actionManager.getAction(ActionKey.REDO).putValue(Action.NAME,
				undoManager.getRedoPresentationName());
	}

	/**
	 * 選択されているノードを消去します。
	 * 
	 */
	public void delete() {
		Set<FlowNode> nodes = new HashSet<FlowNode>(rflow.state
				.getPickedVertices());
		delete_internal(nodes.toArray(new FlowNode[0]));
	}

	private void delete_internal(FlowNode[] nodes) {
		Set<FlowEdge> deletedEdges = new HashSet<FlowEdge>();
		for (FlowNode node : nodes) {
			for (Object o : node.getIncidentEdges())
				deletedEdges.add((FlowEdge) o);

			getGraph().removeVertex(node);
			getPickedState().pick(node, false);
		}

		fireUndoableEditUpdate(new UndoableEditEvent(this, undoableEditFactory
				.createDeleteEdit(nodes, deletedEdges.toArray(new FlowEdge[0]))));
	}

	public void cut() {
		copy_internal(true);
	}

	public void copy() {
		copy_internal(false);
	}

	private void copy_internal(boolean delete) {
		// ViewLayout layout =
		// getSubLayout(getPickedState().getPickedVertices(), null, 0, 0);
		ViewLayout layout = getSubLayout(getPickedState().getPickedVertices());
		if (layout != null) {
			clipedLayout = layout;
			rflow.actionManager.getAction(ActionKey.PASTE).setEnabled(true);
			if (delete)
				delete();
		}
	}

	public void paste() {
		if (clipedLayout != null) {
			Set<FlowNode> set = addSubLayout(clipedLayout, new Point(0, 0));
			getPickedState().clearPickedVertices();
			for (FlowNode node : set)
				getPickedState().pick(node, true);
		}
	}

	public void selectAll() {
		for (Object o : getGraph().getVertices()) {
			getPickedState().pick((ArchetypeVertex) o, true);
		}
	}

	/**
	 * 新しいノードをフローに追加します。
	 * 
	 * @param model
	 *            新しいノードのモデル
	 * @param p
	 *            新しいノードの位置 nullならば自動設定
	 * @param connect
	 *            選択ノードと接続するか
	 * @return 新しく追加されたノード
	 */
	public FlowNode addNode(FlowNode node, Point2D p, boolean connect) {
		getGraph().addVertex(node);
		getGraphLayout().update();

		CompoundEdit edit = new NamedCompoundEdit(Messages
				.getString("Action.new_node")); //$NON-NLS-1$

		edit.addEdit(undoableEditFactory.createAddNodeEdit(node));
		// fireUndoableEditUpdate(new UndoableEditEvent(this,
		// undoableEditFactory.createAddNodeEdit(node)));

		boolean pointWasNull = (p == null);

		FlowNode src = null;// connect source
		if (connect) {
			src = connectNode(node, edit);
			if (p == null && src != null) {
				p = transform(getGraphLayout().getLocation(src));
				p.setLocation(p.getX() + 100, p.getY());
			}
		}

		if (p == null)
			p = new Point(50, 50);
		p = inverseTransform(p);

		if (isUseGrid())
			getGraphLayout().forceMoveOnGrid(node, p.getX(), p.getY());
		else
			getGraphLayout().forceMove(node, p.getX(), p.getY());

		// 位置が設定されていない場合、重なっていたら下にずらす
		if (pointWasNull) {
			while (intersects(node)) {
				p.setLocation(p.getX(), p.getY() + 80);
				getGraphLayout().forceMove(node, p.getX(), p.getY());
			}
		}

		repaint();

		edit.end();
		fireUndoableEditUpdate(new UndoableEditEvent(this, edit));

		return node;
	}

	@SuppressWarnings("unchecked")//$NON-NLS-1$
	public Set<FlowNode> addSubLayout(ViewLayout layout, Point p) {

		if (p == null) {
			int y = getSize().height / 2 - layout.getSize().height / 2;
			p = new Point(0, y);
		}

		// orginal -> newnode map
		Map<FlowNode, FlowNode> nodeMap = new HashMap<FlowNode, FlowNode>();
		FlowGraph subGraph = layout.getGraph();

		// Set<FlowNode> subLayoutNodes = new
		// HashSet<FlowNode>(subGraph.getVertices());
		Set<FlowNode> newNodes = new HashSet<FlowNode>(subGraph.getVertices()
				.size());
		for (Object o : subGraph.getVertices()) {
			FlowNode node = (FlowNode) o;
			FlowNode newNode = node.newInstance();
			getGraph().addVertex(newNode);
			newNodes.add(newNode);
			nodeMap.put(node, newNode);

			getGraphLayout().update();
			// 位置
			Point2D subP = layout.getOffsetLocation(node);
			Point2D newP = new Point2D.Double(subP.getX() + p.getX(), subP.getY() + p.getY());
			newP = inverseTransform(newP);
			getGraphLayout().forceMove(newNode, newP.getX(), newP.getY());
		}

		Set<FlowEdge> newEdges = new HashSet<FlowEdge>(subGraph.getEdges().size());
		for (Object o : subGraph.getEdges()) {
			FlowEdge e = (FlowEdge) o;
			FlowEdge newEdge = (FlowEdge) getGraph().addEdge(
					new FlowEdge(nodeMap.get(e.getSource()), nodeMap.get(e.getDest())));
			newEdges.add(newEdge);
		}

		fireUndoableEditUpdate(new UndoableEditEvent(this, undoableEditFactory
				.createAddSublayoutEdit(newNodes.toArray(new FlowNode[0]),
						newEdges.toArray(new FlowEdge[0]))));

		return newNodes;
	}

	private ViewLayout getSubLayout(final Collection<FlowNode> accept) {
		return getSubLayout(accept, null, new Point(50, 50));
	}

	/**
	 * 指定したノードからサブレイアウトを作成。 nodeMapには、元のノードをキーとして、新しいレイアウトのノード引く エントリーが入れられる。
	 * 
	 * @param accept
	 * @param nodeMap
	 *            nullで自動生成、orgNode -> newNode Map
	 * @return
	 */
	private ViewLayout getSubLayout(final Collection<FlowNode> accept,
			Map<FlowNode, FlowNode> nodeMap, Point space) {
		if (accept.size() < 1)
			return null;

		if (nodeMap == null)
			nodeMap = new HashMap<FlowNode, FlowNode>();

		FlowGraph newGraph = (FlowGraph) getGraph().newInstance();

		for (FlowNode node : accept) {
			FlowNode newNode = node.newInstance();
			newGraph.addVertex(newNode);
			nodeMap.put(node, newNode);
		}

		for (Object o : getGraph().getEdges()) {
			FlowEdge e = (FlowEdge) o;
			if (accept.containsAll(e.getIncidentVertices())) {
				newGraph.addEdge(new FlowEdge(nodeMap.get(e.getSource()),
						nodeMap.get(e.getDest())));
			}
		}

		FlowLayout layout = getGraphLayout();
		Rectangle rect = layout.getVertexRect(accept);

		int gridDist = layout.getGridDistance();
		Point zeroPoint = new Point(rect.x - rect.x % gridDist, rect.y - rect.y
				% gridDist);
		/*
		 * rect.grow(spaceX, spaceY); rect.setFrameFromDiagonal(//ビューへの変換
		 * transform(new Point2D.Double(rect.getMinX(), rect.getMinY())),
		 * transform(new Point2D.Double(rect.getMaxX(), rect.getMaxY())));
		 */

		Map<Vertex, Point2D> locationMap = new HashMap<Vertex, Point2D>();
		for (FlowNode node : accept) {
			Point2D p = layout.getLocation(node);
			locationMap.put(nodeMap.get(node), new Point2D.Double(p.getX()
					- zeroPoint.getX(), p.getY() - zeroPoint.getY()));
		}

		Dimension size;
		Point offset;
		if (space != null) {
			rect.grow(space.x, space.y);
			size = new Dimension((int) rect.getWidth(), (int) rect.getHeight());
			offset = new Point(space.x, space.y);
		} else {
			size = getSize();
			Point2D screenZeroP = inverseTransform(new Point2D.Float(0, 0));
			offset = new Point(zeroPoint.x - (int) screenZeroP.getX(),
					zeroPoint.y - (int) screenZeroP.getY());
		}

		return new ViewLayout(newGraph, locationMap, size, offset);
	}

	/**
	 * vと選択されているノードをエッジでつないで、つないだ先のノードに選択を移す。 つないだソースのノード返す。
	 * エッジの追加ができなかった場合は、nullが返る。 自動的にUndoableEditEvent通知を行います。
	 * 
	 * @param v
	 *            destination node
	 * @return source node. null if failed to connect.
	 */
	private void connectNode(Vertex v) {
		CompoundEdit edit = new NamedCompoundEdit(Messages
				.getString("Action.new_edge")); //$NON-NLS-1$
		connectNode(v, edit);
		edit.end();
		fireUndoableEditUpdate(new UndoableEditEvent(this, edit));
	}

	/**
	 * vと選択されているノードをエッジでつないで、つないだ先のノードに選択を移す。 つないだソースのノード返す。
	 * エッジの追加ができなかった場合は、nullが返る。
	 * 
	 * @param v
	 *            destination node
	 * @return source node. null if failed to connect.
	 */
	private FlowNode connectNode(Vertex v, CompoundEdit edit) {
		if (pickedState.getPickedVertices().size() == 1
				&& !pickedState.isPicked(v)) {
			FlowNode selectedNode = (FlowNode) pickedState.getPickedVertices()
					.iterator().next();
			connectNode(selectedNode, v, edit);
			pickedState.clearPickedVertices();
			pickedState.pick(v, true);
			return selectedNode;
		}
		pickedState.clearPickedVertices();
		pickedState.pick(v, true);
		return null;
	}

	private void connectNode(Vertex src, Vertex dest, CompoundEdit edit) {
		FlowEdge e = new FlowEdge(src, dest);
		if (FlowUtil.makeEdgeAddable(e, this, edit)) {
			getGraph().addEdge(e);
			repaint();
			edit.addEdit(undoableEditFactory
					.createAddEdgeEdit(new FlowEdge[] { e }));
			// fireUndoableEditUpdate(new UndoableEditEvent(this,
			// undoableEditFactory.createAddEdgeEdit(new FlowEdge[]{e})));
		} else {
			JOptionPane
					.showMessageDialog(
							FlowViewer.this,
							Messages.getString("Error.unable_to_add"), Messages.getString("Error.title.error"), JOptionPane.WARNING_MESSAGE); //$NON-NLS-1$ //$NON-NLS-2$
		}
	}

	/**
	 * nodeがほかのノードと重ならないかを調べます
	 * 
	 * @param vertex
	 * @return
	 */
	private boolean intersects(Vertex node) {
		Point2D nodeP = layoutTransform(getGraphLayout().getLocation(node));
		if (nodeP == null)
			return false;
		Shape nodeShape = getRenderer().getVertexShapeFunction().getShape(node);
		nodeShape = AffineTransform.getTranslateInstance(nodeP.getX(),
				nodeP.getY()).createTransformedShape(nodeShape);
		Rectangle nodeRect = nodeShape.getBounds();

		for (Object o : getGraph().getVertices()) {
			Vertex v = (Vertex) o;
			if (node.equals(v))
				continue;
			Shape shape = getRenderer().getVertexShapeFunction().getShape(v);
			Point2D p = layoutTransform(getGraphLayout().getLocation(v));
			if (p == null)
				continue;
			shape = AffineTransform.getTranslateInstance(p.getX(), p.getY())
					.createTransformedShape(shape);
			Rectangle rect = shape.getBounds();

			if (nodeRect.intersects(rect))
				return true;
		}
		return false;
	}

	private class MouseHandler extends MouseAdapter {

		private final OnNodeActionManager onNodeMenu = new OnNodeActionManager();

		private final OnGroundActionManager onGroundMenuHandler = new OnGroundActionManager();

		@Override
		public void mousePressed(MouseEvent e) {
			final FlowViewer viewer = (FlowViewer) e.getSource();

			// キー入力を受け取るためにフォーカスをリクエスト
			viewer.requestFocus();

			// PickedState pickedState = viewer.getPickedState();
			final Point2D p = viewer.inverseViewTransform(e.getPoint());
			final FlowNode v = (FlowNode) viewer.getPickSupport().getVertex(
					p.getX(), p.getY());
			final Edge edge = viewer.getPickSupport().getEdge(p.getX(),
					p.getY());

			/*
			 * ノード上
			 */
			if (v != null) {
				if (SwingUtilities.isLeftMouseButton(e)) {
					// SHIFT+左クリック(ノード上)
					if (e.getModifiers() == (MouseEvent.BUTTON1_MASK | MouseEvent.SHIFT_MASK)) {
						/*
						 * //ピン viewer.getPickedState().addSelectionPin(v);
						 */
						viewer.getPickedState().selectInterval(v);
					}
					// 左ダブルクリック(ノード上)
					else if (e.getClickCount() == 2) {
						NodeModel model = null;
						
						if(v.getModel() instanceof SubflowNodeModel){
							SubflowNodeModel boxNodeModel = (SubflowNodeModel)v.getModel();
							Point screenP = e.getPoint();
							SwingUtilities.convertPointToScreen(screenP, e.getComponent());
							boxNodeModel.showContentsDialog(FlowViewer.this,
									rflow, screenP);
						}else{
							try {
								model = v.getModel().showConfigDialog(rflow.frame);
							} catch (CanceledException e1) {
								return;
							}
							v.setModel(model);
						}
					}
					// ALT+左クリック(ノード上)
					else if (e.getModifiers() == (MouseEvent.BUTTON1_MASK | MouseEvent.ALT_MASK)) {
						connectNode(v);
					}
					/*
					 * else{ viewer.getPickedState().clearPinNodes(); }
					 */
				}
				// 中クリック(ノード上)
				else if (SwingUtilities.isMiddleMouseButton(e)) {
					connectNode(v);
				}
				// 右クリック(ノード上)
				else if (SwingUtilities.isRightMouseButton(e)) {
					/*
					 * if(viewer.getPickedState().isPinFlowSelected() &&
					 * viewer.getPickedState().getSelectionNodeFlow().contains(v)) {
					 * onNodeMenu.showFlowSelectedMenu(viewer, viewer, v,
					 * (int)p.getX(), (int)p.getY()); } else {
					 */
					onNodeMenu.showNormalMenu(viewer, viewer, v,
							(int) p.getX(), (int) p.getY());
					// }
				}
			}
			/*
			 * ノード以外
			 */
			else {
				/*
				 * エッジ上
				 */
				if (edge != null) {
					if (SwingUtilities.isRightMouseButton(e)) {

						JMenuItem delete = new JMenuItem(Messages
								.getString("Action.remove_edge")); //$NON-NLS-1$
						delete.addActionListener(new ActionListener() {
							public void actionPerformed(ActionEvent e) {
								viewer.getGraph().removeEdge(edge);
								viewer.repaint();

								fireUndoableEditUpdate(new UndoableEditEvent(
										this,
										undoableEditFactory.createDeleteEdit(
												null,
												new FlowEdge[] { (FlowEdge) edge })));
							}
						});

						JPopupMenu popup = new JPopupMenu();
						popup.add(delete);
						popup.show(viewer, (int) p.getX(), (int) p.getY());
					}
				}
				/*
				 * 何もないところ
				 */
				else if (SwingUtilities.isLeftMouseButton(e)) {
					// 地面ダブルクリック
					if (e.getClickCount() == 2) {
						NodeModel model;
						try {
							model = NodeModelManager.getModelController(
									SentenceNodeModel.class).createNewModel();
						} catch (CanceledException e1) {
							return;
						}
						addNode(new FlowNode(model), e.getPoint(), true);
					}/*
						 * else{ viewer.getPickedState().clearPinNodes(); }
						 */
				} else if (SwingUtilities.isRightMouseButton(e)) {
					// 何もないところを右クリック
					onGroundMenuHandler.showPopup(FlowViewer.this, (int) e
							.getPoint().getX(), (int) e.getPoint().getY());
				}

			}
		}
	}

	private enum OnGroundActionKey {
		CREATE_SENTENCE_NODE, 
		CREATE_FREE_NODE,
		CREATE_SUBFLOW_NODE,
		INSERT_SOURCE, 
		SELECT_ALL;
	}

	private class OnGroundActionManager extends
			AbstractActionManager<OnGroundActionKey> {

		private FlowViewer viewer;

		private Point p;

		// private Action[] actions;

		OnGroundActionManager() {
			putAction(OnGroundActionKey.CREATE_SENTENCE_NODE,
					new AbstractAction(Messages
							.getString("Action.new_sentence_node")) { //$NON-NLS-1$
						public void actionPerformed(ActionEvent e) {
							createModel(NodeModelManager
									.getModelController(SentenceNodeModel.class));
						}
					});

			putAction(OnGroundActionKey.CREATE_FREE_NODE, new AbstractAction(
					Messages.getString("Action.new_free_node")) { //$NON-NLS-1$
						public void actionPerformed(ActionEvent e) {
							createModel(NodeModelManager
									.getModelController(FreeNodeModel.class));
						}
					});
			
			putAction(OnGroundActionKey.CREATE_SUBFLOW_NODE, new AbstractAction("サブフローノードの作成"){
				public void actionPerformed(ActionEvent e) {
					createModel(NodeModelManager.getModelController(SubflowNodeModel.class));
				}
			});

			putAction(OnGroundActionKey.INSERT_SOURCE, new AbstractAction(
					Messages.getString("Action.insert_source")) { //$NON-NLS-1$
						public void actionPerformed(ActionEvent e) {
							ViewLayout layout = RflowUtil
									.importSource(rflow.frame);
							if (layout == null)
								return;

							Point offset = ViewLayout
									.getDefaultFirstnodePoint();
							Point point = (Point) p.clone();
							point.translate(-offset.x, -offset.y);
							viewer.addSubLayout(layout, point);
						}
					});

			putAction(OnGroundActionKey.SELECT_ALL, new AbstractAction(Messages
					.getString("Action.select_all")) { //$NON-NLS-1$
						public void actionPerformed(ActionEvent e) {
							selectAll();
						}
					});
		}

		private void createModel(NodeModelController controller) {
			NodeModel model;
			try {
				model = controller.createNewModel();
			} catch (CanceledException e1) {
				return;
			}
			addNode(new FlowNode(model), p, true);
		}

		public void showPopup(FlowViewer viewer, int x, int y) {
			this.viewer = viewer;
			this.p = new Point(x, y);

			JPopupMenu popup = new JPopupMenu();
			popup.add(getAction(OnGroundActionKey.CREATE_SENTENCE_NODE));
			popup.add(getAction(OnGroundActionKey.CREATE_FREE_NODE));
			popup.add(getAction(OnGroundActionKey.INSERT_SOURCE));
			popup.addSeparator();
			popup.add(getAction(OnGroundActionKey.SELECT_ALL));

			popup.show(viewer, x, y);
		}

	}

	private enum OnNodeActionKey {
		EXECUTE,
		EXECUTE_CLEAR,
		// EXECUTE_LOCAL,
		// EXECUTE_FLAG,
		COPY_SCRIPT, 
		REMOVE_NODE, 
		TAKE_OUT, 
		BRANCH, 
		REMOVE_EDGE, 
		SELECT_DESCENDANT// ,
		// CREATE_FUNCTION,
		// CREATE_SUBFLOW,
		// SAVE_AS_PARTS
		;

	}

	private class OnNodeActionManager extends
			AbstractActionManager<OnNodeActionKey> {

		private FlowViewer viewer;

		private FlowNode node;

		// private Point p;

		OnNodeActionManager() {

			putAction(OnNodeActionKey.EXECUTE, new AbstractAction(Messages
					.getString("Action.run")) { //$NON-NLS-1$
						public void actionPerformed(ActionEvent e) {
							executeRest(node);
						}
					});

			putAction(OnNodeActionKey.EXECUTE_CLEAR, new AbstractAction(
					Messages.getString("Action.run_clear")) { //$NON-NLS-1$
						public void actionPerformed(ActionEvent e) {
							execute(node, true);
						}
					});

			/*
			 * putAction(OnNodeActionKey.EXECUTE_LOCAL, new
			 * AbstractAction("ローカルで実行"){ public void
			 * actionPerformed(ActionEvent e) { executer.excuteInLocal(node);
			 * viewer.repaint(); } });
			 */

			/*
			 * putAction(OnNodeActionKey.EXECUTE_FLAG, new
			 * AbstractAction("フラグから実行"){ public void
			 * actionPerformed(ActionEvent e) { executer.excuteRest(node);
			 * getPickedState().setFlagNode(node); viewer.repaint(); } });
			 */

			putAction(OnNodeActionKey.COPY_SCRIPT, new AbstractAction(Messages
					.getString("Action.copy_script")) { //$NON-NLS-1$
						public void actionPerformed(ActionEvent e) {
							copyScript(node);
						}
					});

			putAction(OnNodeActionKey.REMOVE_NODE, new AbstractAction(Messages
					.getString("Action.remove_node")) { //$NON-NLS-1$
						public void actionPerformed(ActionEvent e) {
							delete_internal(getNodeSelected());
						}
					});

			putAction(OnNodeActionKey.TAKE_OUT, new AbstractAction(Messages
					.getString("Action.take_out")) { //$NON-NLS-1$
						public void actionPerformed(ActionEvent e) {
							FlowNode[] nodes = getNodeSelected();
							CompoundEdit edit = new NamedCompoundEdit(getValue(
									Action.NAME).toString());
							if (nodes.length == 1 && node.outDegree() < 1) {
								if (!node.getInEdges().isEmpty()) {
									FlowEdge edge = (FlowEdge) node
											.getInEdges().iterator().next();
									getGraph().removeEdge(edge);
									edit.addEdit(undoableEditFactory
											.createDeleteEdit(null,
													new FlowEdge[] { edge }));
									edit.end();
									fireUndoableEditUpdate(new UndoableEditEvent(
											FlowViewer.this, edit));
								}
							} else {
								FlowUtil.takeOut(getNodeSelected(), false,
										undoableEditFactory, edit);
								viewer.getPickedState().clearPickedVertices();
								edit.end();
								fireUndoableEditUpdate(new UndoableEditEvent(
										FlowViewer.this, edit));
							}
						}
					});

			putAction(OnNodeActionKey.BRANCH, new AbstractAction(Messages
					.getString("Action.branch")) { //$NON-NLS-1$
						public void actionPerformed(ActionEvent e) {
							CompoundEdit edit = new NamedCompoundEdit(getValue(
									Action.NAME).toString());
							FlowUtil.takeOut(getNodeSelected(), true,
									undoableEditFactory, edit);
							edit.end();
							fireUndoableEditUpdate(new UndoableEditEvent(
									FlowViewer.this, edit));
						}
					});

			putAction(OnNodeActionKey.REMOVE_EDGE, new AbstractAction(Messages
					.getString("Action.remove_edge")) { //$NON-NLS-1$
						public void actionPerformed(ActionEvent e) {
							if (!node.getInEdges().isEmpty()) {
								FlowEdge edge = (FlowEdge) node.getInEdges()
										.iterator().next();
								getGraph().removeEdge(edge);
								fireUndoableEditUpdate(new UndoableEditEvent(
										FlowViewer.this,
										undoableEditFactory.createDeleteEdit(
												null, new FlowEdge[] { edge })));
							}
						}
					});

			putAction(OnNodeActionKey.SELECT_DESCENDANT, new AbstractAction(
					Messages.getString("Action.select_children")) { //$NON-NLS-1$
						public void actionPerformed(ActionEvent e) {
							FlowNode[] selected = getNodeSelected();
							viewer.getPickedState().clearPickedVertices();
							for (FlowNode node : FlowUtil.getChildren(selected)) {
								viewer.getPickedState().pick(node, true);
							}
							viewer.repaint();
						}
					});

			/*
			 * putAction(OnNodeActions.CREATE_FUNCTION, new
			 * AbstractAction("関数を作成"){ public void actionPerformed(ActionEvent
			 * e) { if(!viewer.getPickedState().isPinFlowSelected())return;
			 * 
			 * String command = ""; //$NON-NLS-1$ List<FlowNode> stream =
			 * viewer.getPickedState().getSelectionNodeFlow(); for(FlowNode n :
			 * stream) command += n.getModel().getCommand();
			 * //TODO整形にRのparseを使うことを検討 command = command.replaceAll("\n",
			 * "\n\t"); //$NON-NLS-1$ //$NON-NLS-2$ command = "\t" + command;
			 * //$NON-NLS-1$ if(command.endsWith("\t"))command =
			 * command.substring(0, command.length()-1); //$NON-NLS-1$
			 * 
			 * command = "function()"+Constants.NEWLINE+ //$NON-NLS-1$
			 * "{"+Constants.NEWLINE+ //$NON-NLS-1$ command+
			 * "}"+Constants.NEWLINE; //$NON-NLS-1$
			 * 
			 * REXP globals = rflow.R.eval("findGlobals("+command+",
			 * merge=FALSE)"); //$NON-NLS-1$ //$NON-NLS-2$ if(globals != null &&
			 * globals.asList() != null){ RList list = globals.asList();
			 * String[] variables = list.at("variables").asStringArray();
			 * //$NON-NLS-1$ if(variables != null && variables.length > 0){
			 * String args = ""; //$NON-NLS-1$ for(String s : variables) args +=
			 * s + ", "; //$NON-NLS-1$ args = args.substring(0,
			 * args.length()-2);
			 * 
			 * command = command.replace("function()", "function("+args+")");
			 * //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ } }
			 * 
			 * FreeNodeModel model = new FreeNodeModel();
			 * model.setTitle("new_function"); //$NON-NLS-1$ model.command =
			 * command; try { model =
			 * (FreeNodeModel)NodeModelManager.getModelController(model.getClass()).getNodeModel(model); }
			 * catch (CanceledException ex) { return; } Point pp =
			 * (Point)p.clone(); pp.translate(0, 50); addNode(new
			 * FlowNode(model), pp, false);
			 * //rflow.explorerPanel.addUserNode(model); } });
			 */

			/*
			 * putAction(OnNodeActions.CREATE_SUBFLOW, new
			 * AbstractAction("取り出し"){ public void actionPerformed(ActionEvent
			 * e) { if(!viewer.getPickedState().isPinFlowSelected())return; List<FlowNode>
			 * stream = getPickedState().getSelectionNodeFlow(); Map<FlowNode,
			 * FlowNode> nodeMap = new HashMap<FlowNode, FlowNode>();
			 * 
			 * ViewLayout viewLayout = getSubLayout(stream, nodeMap, 150, 100);
			 * 
			 * FlowNode in = stream.get(0); FlowNode out =
			 * stream.get(stream.size()-1); BoxNodeModel model = new
			 * BoxNodeModel(viewLayout, nodeMap.get(in), nodeMap.get(out));
			 * FlowNode newNode = addNode(new FlowNode(model),
			 * transform(getGraphLayout().getCentroid(stream)), false); List<FlowEdge>
			 * addEdges = new ArrayList<FlowEdge>();
			 * if(in.getPredecessors().size() == 1){ FlowEdge e1 =
			 * (FlowEdge)getGraph().addEdge(new
			 * FlowEdge((FlowNode)in.getPredecessors().iterator().next(),
			 * newNode)); addEdges.add(e1); } if(out.getSuccessors().size() ==
			 * 1){ FlowEdge e2 = (FlowEdge)getGraph().addEdge(new
			 * FlowEdge(newNode,
			 * (FlowNode)out.getSuccessors().iterator().next()));
			 * addEdges.add(e2); } if(addEdges.size() > 0){
			 * fireUndoableEditUpdate(new UndoableEditEvent(FlowViewer.this,
			 * undoableEditFactory.createAddEdgeEdit(addEdges.toArray(new
			 * FlowEdge[0])))); }
			 * 
			 * //System.out.println(g); delete_internal(stream.toArray(new
			 * FlowNode[0])); //getGraph().removeVertices(new HashSet<FlowNode>(stream)); }
			 * });
			 */

			/*
			 * putAction(OnNodeActions.SAVE_AS_PARTS, new
			 * AbstractAction("選択部分をパーツ登録"){ @SuppressWarnings("unchecked")
			 * //$NON-NLS-1$ public void actionPerformed(ActionEvent e) { Set<FlowNode>
			 * picked = viewer.getPickedState().getPickedVertices();
			 * if(picked.size() < 1)return;
			 * 
			 * ViewLayout subLayout = getSubLayout(picked);
			 * 
			 * rflow.explorer.getPartsExplorerPanel().addParts("NEW_PARTS",
			 * subLayout); viewer.getPickedState().clearPickedVertices(); } });
			 */
		}

		/**
		 * クリックされたノードが選択されていたら選択ノード全体 選択されていなかったらクリックされたノードのみを返す
		 * 
		 * @return
		 */
		@SuppressWarnings("unchecked")//$NON-NLS-1$
		private FlowNode[] getNodeSelected() {
			if (viewer.getPickedState().isPicked(node)) {
				Set<FlowNode> picked = viewer.getPickedState()
						.getPickedVertices();
				return picked.toArray(new FlowNode[0]);
			} else {
				return new FlowNode[] { node };
			}
		}

		public void showNormalMenu(Component invoker, FlowViewer viewer,
				FlowNode node, int x, int y) {
			this.viewer = viewer;
			this.node = node;

			JPopupMenu popup = new JPopupMenu();

			popup.add(getAction(OnNodeActionKey.EXECUTE));
			popup.add(getAction(OnNodeActionKey.EXECUTE_CLEAR));
			popup.add(getAction(OnNodeActionKey.COPY_SCRIPT));
			popup.addSeparator();
			popup.add(getAction(OnNodeActionKey.REMOVE_NODE));
			popup.add(getAction(OnNodeActionKey.TAKE_OUT));
			popup.add(getAction(OnNodeActionKey.BRANCH));
			popup.add(getAction(OnNodeActionKey.SELECT_DESCENDANT));
			popup.addSeparator();
			popup.add(getAction(OnNodeActionKey.REMOVE_EDGE));

			popup.show(invoker, x, y);
		}


	}

	private class DropTargetHandler extends DropTargetAdapter {

		@SuppressWarnings("unchecked")//$NON-NLS-1$
		public void drop(DropTargetDropEvent dtde) {
			Transferable trans = dtde.getTransferable();
			// 関数ノード：関数名
			if (trans.isDataFlavorSupported(FunctionNodeFlavor.getInstance())) {
				try {
					// String function =
					// trans.getTransferData(DataFlavor.stringFlavor).toString();
					FunctionNode[] nodes = (FunctionNode[]) trans
							.getTransferData(FunctionNodeFlavor.getInstance());
					FunctionNode node = nodes[0];
					if (nodes.length > 1 | !node.hasFunction()) {
						dtde.rejectDrop();
						dtde.dropComplete(false);
						return;
					}

					NodeModel nodeModel = new SentenceNodeModel();

					try {
						nodeModel.setCommand(node.getFunction() + "()"); //$NON-NLS-1$
					} catch (IncompatibleCommandException e) {
						// TODO 再読み込み検討
						e.printStackTrace();
					}
					// nodeModel.function = function;

					NodeModel model = NodeModelManager.getModelController(
							nodeModel.getClass()).getNodeModel(nodeModel);

					Point p = dtde.getLocation();
					addNode(new FlowNode(model), p, true);
				} catch (UnsupportedFlavorException e) {
					dtde.dropComplete(false);
				} catch (IOException e) {
					dtde.dropComplete(false);
				} catch (CanceledException e) {
				}
			}
			// レイアウト：パーツ
			else if (trans
					.isDataFlavorSupported(LayoutDataFlavor.getInstance())) {
				try {
					ViewLayout layout = (ViewLayout) trans
							.getTransferData(LayoutDataFlavor.getInstance());
					// System.out.println(layout.getGraph());
					Point p = dtde.getLocation();
					p.translate(-layout.getSize().width / 2,
							-layout.getSize().height / 2);
					addSubLayout(layout, p);
					removeDragImage();
				} catch (UnsupportedFlavorException e) {
					dtde.dropComplete(false);
				} catch (IOException e) {
					dtde.dropComplete(false);
				}
			}
			// ファイル:読み込みノード作成
			else if (trans.isDataFlavorSupported(DataFlavor.javaFileListFlavor)) {
				dtde.acceptDrop(DnDConstants.ACTION_LINK);
				List<File> list;
				try {
					list = (List<File>) trans
							.getTransferData(DataFlavor.javaFileListFlavor);
				} catch (Exception e) {
					dtde.rejectDrop();
					dtde.dropComplete(false);
					return;
				}

				if (list.size() > 1) {
					dtde.dropComplete(false);
					return;
				}

				File file = list.get(0);

				if (file.getName().endsWith(Constants.RFLOW_FILE_EXTENSION)) {
					// .rflowファイルの場合
					if (rflow.getWorkingFile() != null
							&& rflow.getWorkingFile().equals(file))
						return;
					if (!rflow.actionManager.askSave())
						return;
					rflow.actionManager.open(file, false);
				} else {
					NodeModel model = NodeModelFactory.createNodeModel(file,
							Prefs.WORKING_DIRECTORY);
					if (model == null) {
						dtde.dropComplete(false);
						return;
					}

					try {
						model = model.showConfigDialog(rflow.frame);
					} catch (CanceledException e1) {
						dtde.dropComplete(true);
						return;
					}

					addNode(new FlowNode(model), dtde.getLocation(), false);
					dtde.dropComplete(true);
				}
			}
			dtde.dropComplete(false);
		}

		ImagePaintable partsDragImage;

		@SuppressWarnings("unchecked")//$NON-NLS-1$
		@Override
		public void dragEnter(DropTargetDragEvent dtde) {
			Transferable trans = dtde.getTransferable();
			if (trans.isDataFlavorSupported(LayoutDataFlavor.getInstance())
					&& trans.isDataFlavorSupported(DataFlavor.imageFlavor)) {
				try {
					Image image = (Image) trans
							.getTransferData(DataFlavor.imageFlavor);
					partsDragImage = new ImagePaintable(image);
					partsDragImage.setPoint(dtde.getLocation());
					FlowViewer.this.addPostRenderPaintable(partsDragImage);
					FlowViewer.this.repaint();
				} catch (UnsupportedFlavorException e) {
					dtde.rejectDrag();
				} catch (IOException e) {
					dtde.rejectDrag();
				}
			} else if (trans
					.isDataFlavorSupported(DataFlavor.javaFileListFlavor)) {
				List<File> list;
				try {
					list = (List<File>) trans
							.getTransferData(DataFlavor.javaFileListFlavor);
				} catch (Exception e) {
					dtde.rejectDrag();
					return;
				}

				if (list.size() > 1) {
					dtde.rejectDrag();
					return;
				}

				File file = list.get(0);
				if (!file.getName().endsWith(Constants.RFLOW_FILE_EXTENSION)
						&& !NodeModelFactory.isModelCreationSupported(file))
					dtde.rejectDrag();
			}
		}

		@Override
		public void dragOver(DropTargetDragEvent dtde) {
			if (partsDragImage != null) {
				partsDragImage.setPoint(dtde.getLocation());
				FlowViewer.this.repaint();
			}
		}

		@Override
		public void dragExit(DropTargetEvent dte) {
			removeDragImage();
			FlowViewer.this.repaint();
		}

		private void removeDragImage() {
			if (partsDragImage != null) {
				FlowViewer.this.removePostRenderPaintable(partsDragImage);
				partsDragImage.flushImage();
				partsDragImage = null;
			}
		}
	}

	private class GraphEventHandler implements GraphEventListener {

		public void edgeAdded(GraphEvent event) {
			update();
		}

		public void edgeRemoved(GraphEvent event) {
			update();
		}

		public void vertexAdded(GraphEvent event) {
			FlowNode node = (FlowNode) event.getGraphElement();
			updateGroup(node.getModel());
			update();
		}

		public void vertexRemoved(GraphEvent event) {
			update();
		}

		private void update() {
			rflow.setDirty(true);
			repaint();
		}
	}

	private static class ToolTipHandler implements ToolTipFunction {

		public String getToolTipText(Vertex v) {
			FlowNode node = (FlowNode) v;
			return convertToHTML(node.getModel().getProperty().getComment());
		}

		public String getToolTipText(Edge e) {
			return null;
		}

		public String getToolTipText(MouseEvent event) {
			return null;
		}

		private String convertToHTML(String s) {
			if (s == null)
				return null;
			String html = "<html>" + s + "</html>"; //$NON-NLS-1$ //$NON-NLS-2$
			html = html.replaceAll("\\n", "<br>"); //$NON-NLS-1$ //$NON-NLS-2$
			return html;
		}
	}

	public FlowExecuter getExecuter() {
		return executer;
	}

	/**
	 * ビューでのノードの位置
	 */
	public Map<Vertex, Point2D> getLocationMap() {
		Map<Vertex, Point2D> locationMap = new HashMap<Vertex, Point2D>();

		FlowLayout layout = getGraphLayout();
		locationMap.clear();

		for (Object o : getGraph().getVertices()) {
			FlowNode node = (FlowNode) o;
			Point2D p = transform(layout.getLocation(node));
			locationMap.put(node, p);
		}

		return locationMap;
	}

	@SuppressWarnings("unchecked")
	public ViewLayout getViewLayout() {
		if(getGraph().getVertices().size() == 0){
			return new ViewLayout(new FlowGraph(), new HashMap<Vertex, Point2D>(), 
					getSize(), new Point());
		}
		return getSubLayout(getGraph().getVertices(), null, null);
	}

	public void paintToGraphics(Graphics2D g) {
		renderGraph(g);
	}

	private class ImagePaintable implements Paintable {

		private Image image;

		// private ImageIcon icon;
		private Point p = new Point();

		public ImagePaintable(Image image) {
			super();
			this.image = image;

			// icon = new ImageIcon(image);
		}

		public void paint(Graphics g) {

			// icon.paintIcon(StreamViewer.this, g, p.x-icon.getIconWidth()/2,
			// p.y-icon.getIconHeight()/2);
			ImageObserver observer = FlowViewer.this;
			g.drawImage(image, p.x - image.getWidth(observer) / 2, p.y
					- image.getHeight(observer) / 2, observer);
		}

		public void setPoint(Point p) {
			this.p = p;
		}

		public void flushImage() {
			image.flush();// 必要？
		}

		public boolean useTransform() {
			return false;
		}

	}

	@SuppressWarnings("unchecked")//$NON-NLS-1$
	protected void renderGraph(Graphics2D g2d) {

		Layout layout = model.getGraphLayout();

		g2d.setRenderingHints(renderingHints);

		// the size of the VisualizationViewer
		Dimension d = getSize();

		// clear the offscreen image
		g2d.setColor(getBackground());
		g2d.fillRect(0, 0, d.width, d.height);

		AffineTransform oldXform = g2d.getTransform();
		AffineTransform newXform = new AffineTransform(oldXform);
		newXform.concatenate(viewTransformer.getTransform());

		g2d.setTransform(newXform);

		// if there are preRenderers set, paint them
		for (Iterator iterator = preRenderers.iterator(); iterator.hasNext();) {
			Paintable paintable = (Paintable) iterator.next();
			if (paintable.useTransform()) {
				paintable.paint(g2d);
			} else {
				g2d.setTransform(oldXform);
				paintable.paint(g2d);
				g2d.setTransform(newXform);
			}
		}

		locationMap.clear();

		// paint all the edges
		//
		getRenderer().clearArrowOnlyEdges();
		try {
			for (Iterator iter = layout.getGraph().getEdges().iterator(); iter
					.hasNext();) {
				Edge e = (Edge) iter.next();
				Vertex v1 = (Vertex) e.getEndpoints().getFirst();
				Vertex v2 = (Vertex) e.getEndpoints().getSecond();

				Point2D p = (Point2D) locationMap.get(v1);
				if (p == null) {

					p = layout.getLocation(v1);
					p = layoutTransformer.transform(p);
					locationMap.put(v1, p);
				}
				Point2D q = (Point2D) locationMap.get(v2);
				if (q == null) {
					q = layout.getLocation(v2);
					q = layoutTransformer.transform(q);
					locationMap.put(v2, q);
				}

				if (p != null && q != null) {
					renderer.paintEdge(g2d, e, (int) p.getX(), (int) p.getY(),
							(int) q.getX(), (int) q.getY());
				}
			}
		} catch (ConcurrentModificationException cme) {
			repaint();
		}

		// paint all the vertices label space
		try {
			for (Iterator iter = layout.getGraph().getVertices().iterator(); iter
					.hasNext();) {

				Vertex v = (Vertex) iter.next();
				Point2D p = (Point2D) locationMap.get(v);
				if (p == null) {
					p = layout.getLocation(v);
					p = layoutTransformer.transform(p);
					locationMap.put(v, p);
				}
				if (p != null) {
					getRenderer().paintVertexSpaces(g2d, v, (int) p.getX(),
							(int) p.getY());
				}
			}
		} catch (ConcurrentModificationException cme) {
			repaint();
		}

		// paint all the vertices
		try {
			for (Iterator iter = layout.getGraph().getVertices().iterator(); iter
					.hasNext();) {

				Vertex v = (Vertex) iter.next();
				Point2D p = (Point2D) locationMap.get(v);
				if (p == null) {
					p = layout.getLocation(v);
					p = layoutTransformer.transform(p);
					locationMap.put(v, p);
				}
				if (p != null) {
					renderer.paintVertex(g2d, v, (int) p.getX(), (int) p.getY());
				}
			}
		} catch (ConcurrentModificationException cme) {
			repaint();
		}

		// arrow only edge
		try {
			for (Edge e : getRenderer().getArrowOnlyEdges()) {
				Vertex v1 = (Vertex) e.getEndpoints().getFirst();
				Vertex v2 = (Vertex) e.getEndpoints().getSecond();

				Point2D p = (Point2D) locationMap.get(v1);
				if (p == null) {
					p = layout.getLocation(v1);
					p = layoutTransformer.transform(p);
					locationMap.put(v1, p);
				}
				Point2D q = (Point2D) locationMap.get(v2);
				if (q == null) {
					q = layout.getLocation(v2);
					q = layoutTransformer.transform(q);
					locationMap.put(v2, q);
				}

				if (p != null && q != null) {
					getRenderer().paintArrowOnlyEdge(g2d, e, (int) p.getX(),
							(int) p.getY(), (int) q.getX(), (int) q.getY());
				}
			}
		} catch (ConcurrentModificationException cme) {
			repaint();
		}

		// if there are postRenderers set, do it
		for (Iterator iterator = postRenderers.iterator(); iterator.hasNext();) {
			Paintable paintable = (Paintable) iterator.next();
			if (paintable.useTransform()) {
				paintable.paint(g2d);
			} else {
				g2d.setTransform(oldXform);
				paintable.paint(g2d);
				g2d.setTransform(newXform);
			}
		}
		g2d.setTransform(oldXform);
	}

}
