package com.ef_prime.rflow.flow;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Map;

import com.ef_prime.rflow.flow.model.FlowEdge;
import com.ef_prime.rflow.flow.model.FlowGraph;
import com.ef_prime.rflow.flow.model.FlowNode;
import com.ef_prime.rflow.node.NodeModel;

import edu.uci.ics.jung.graph.Vertex;

public class ViewLayout {
	
	private static final Point DEFAULT_FIRSTNODE_POINT = new Point(80, 80);
	
	private FlowGraph graph;
	private Map<Vertex, Point2D> locationMap;
	private Dimension size;
	private Point offset;
	
	public ViewLayout(FlowGraph graph, Map<Vertex, Point2D> locationMap, Dimension size, Point offset){
		this.graph  = graph;
		this.locationMap = locationMap;
		this.size = size;
		this.offset = offset;
	}
	
	public ViewLayout(NodeModel[] models){
		
		graph = new FlowGraph();
		FlowNode lastNode = null;
		locationMap = new HashMap<Vertex, Point2D>();
		int hSpace = DEFAULT_FIRSTNODE_POINT.x;
		int vSpace = DEFAULT_FIRSTNODE_POINT.y;
		
		int x = hSpace;
		for(NodeModel model : models){
			FlowNode node = new FlowNode(model);
			graph.addVertex(node);
			locationMap.put(node, new Point(x, vSpace));
			x += hSpace;
			
			if(lastNode != null)
				graph.addEdge(new FlowEdge(lastNode, node));
			
			lastNode = node;
		}
		
		size = new Dimension((models.length+1)*hSpace, vSpace*2);
		offset = new Point();
	}

	public FlowGraph getGraph() {
		return graph;
	}

	public Map<Vertex, Point2D> getLocationMap() {
		return locationMap;
	}

	public Dimension getSize() {
		return size;
	}
	
	public Point getOffset() {
		return offset;
	}
	
	public Point2D getLocation(Vertex node){
		return locationMap.get(node);
	}
	
	public Point2D getOffsetLocation(Vertex node){
		Point2D p = locationMap.get(node);
		return new Point2D.Double(p.getX()+offset.x, p.getY()+offset.y);
	}

	public static Point getDefaultFirstnodePoint() {
		return (Point)DEFAULT_FIRSTNODE_POINT.clone();
	}
}
