package com.ef_prime.rflow.node;

import java.awt.Shape;
import java.awt.Window;

import javax.swing.Icon;

import com.ef_prime.rflow.Constants;
import com.ef_prime.rflow.exception.CanceledException;
import com.ef_prime.rflow.exception.IncompatibleCommandException;
import com.ef_prime.rflow.flow.rendering.NodeShape;
import com.ef_prime.rflow.group.AppearanceGroupManager;
import com.ef_prime.rflow.util.ObjectUtilities;

public abstract class AbstractNodeModel implements NodeModel{
	
	private String title;
	private NodeProperty property = new NodeProperty();
	
	public AbstractNodeModel(){}
	
	/**
	 * コピー用コンストラクタ
	 * @param model
	 */
	protected AbstractNodeModel(NodeModel model){
		if(model.isTitle())
			setTitle(model.getTitle());
		
		property = model.getProperty().copy();
	}
	
	public final String getCommandNewline() {
		String c = getCommand();
		if(c.length() < 1 || c.charAt(c.length()-1) != Constants.NEWLINE)
			c += Constants.NEWLINE;
		return c;
	}
	
	public final String getTitle() {
		if(title == null)
			return getDefaultTitle();
		return title;
	}
	
	protected String getDefaultTitle(){
		return ""; //$NON-NLS-1$
	}
	
	public final boolean isTitle(){
		return title != null;
	}
	
	public void setTitle(String title) {
		this.title = title;
	}
	
	public final NodeProperty getProperty() {
		return property;
	}
	
	public final void setProperty(NodeProperty property) {
		if(property == null)
			throw new IllegalArgumentException("property can't be null"); //$NON-NLS-1$
		this.property = property;
	}
	
	public Icon getIcon() {
		return null;
	}
	
	public Icon getSmallIcon() {
		return getNodeShape().getSmallIcon();
	}
	
	public Shape getShape() {
		return getNodeShape().getShape();
	}
	
	protected NodeShape getNodeShape(){
		return NodeShape.RECTANGLE;
	}
	
	public Shape getGlassShape() {
		return NodeShape.getGlassShape();
	}
	
	public void update(AppearanceGroupManager groupManager) {
	}
	
	public boolean isEditable() {
		return false;
	}
	
	public void setCommand(String command) throws IncompatibleCommandException{
		if(!isEditable())
			throw new UnsupportedOperationException("cannot invoke this method if not editable"); //$NON-NLS-1$
	}
	

	public NodeModel showConfigDialog(Window parent) throws CanceledException {
		//return getController().getNodeModel(this);
		return NodeModelManager.getModelController(this.getClass()).getNodeModel(this);
	}
	
	
	abstract public NodeModel copy();
	
	@Override
	public int hashCode() {
		return getCommand().hashCode();
	}
	
	public boolean equals(Object obj) {
		if (obj instanceof NodeModel) {
			NodeModel model = (NodeModel) obj;
			return ObjectUtilities.equals(getTitle(), model.getTitle())
				&&
				ObjectUtilities.equals(getCommand(), model.getCommand())
				&&
				ObjectUtilities.equals(getProperty(), model.getProperty());
		}
		return false;
	}
}
