package com.ef_prime.rflow.node;

import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.HashSet;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.border.TitledBorder;
import javax.swing.filechooser.FileFilter;

import org.rosuda.JRI.REXP;

import com.ef_prime.rflow.Constants;
import com.ef_prime.rflow.Prefs;
import com.ef_prime.rflow.RCommand;
import com.ef_prime.rflow.exception.InvalidOptionException;
import com.ef_prime.rflow.util.FileChooserUtility;
import com.ef_prime.rflow.util.FileUtility;

public class DefaultArgComponents {
	
	private static int DEFAULT_HEIGHT = 20;
	
	public static ACFreeInput createSymbolInput(String label){
		ACFreeInput fi = new ACFreeInput("", label, false);
		fi.field.setPreferredSize(new Dimension(150, DEFAULT_HEIGHT));
		fi.field.setMaximumSize(fi.field.getPreferredSize());
		return fi;
	}
	
	public static ACFreeInput createFreeInput(String param, String label, boolean quote){
		return new ACFreeInput(param, label, quote);
	}
	
	public static ACFileInput createFileInput(String param, String label, boolean quote){
		return new ACFileInput(param, label, quote);
	}
	
	public static ACRadioButtons createRadioButton(String param, String title,
			String[] labels, String[] values, boolean quote)
	{
		return new ACRadioButtons(param, title, labels, values, quote);
	}

	public static ACComboBox createComboBox(String param, String title,
			String[] labels, String[] values)
	{
		return new ACComboBox(param, title, labels, values);
	}
	
	public static ACList createList(String param, String title,
			String[] labels, String[] values, boolean singleSelection)
	{
		return new ACList(param, title, labels, values, singleSelection);
	}
	
	public static ACComboBox createDataFrameChooser(String param){
		String[] dataframes = RCommand.getObjectsFiltered(RCommand.FILTER_DATAFRAME);
		if(dataframes == null)dataframes = new String[]{"NO DATAFRAME"};
		return createComboBox(param, "データ", dataframes, dataframes);
	}
	
	public static ArgComponent createSimpleFormulaEditor(String paramFormula, String paramData, 
			String title, String title1, String title2)
	{
		ArgComponent ac = ArgComponent.createNoParamArgumentComponent();
		ac.setLayout(new BoxLayout(ac, BoxLayout.Y_AXIS));
		ac.setBorder(new TitledBorder(title));
		
		ACComboBox data = createDataFrameChooser(paramData);
		ACSimpleFormula formula = new ACSimpleFormula(paramFormula, title1, title2, data.comboBox);
		
		ac.add(data);
		ac.add(formula);
		
		ac.addArgumentComponent(data);
		ac.addArgumentComponent(formula);
		
		ac.setMaximumSize(ac.getPreferredSize());
		
		return ac;
	}
	
	public static ACCheckBox createCheckBox(String param, String label){
		return new ACCheckBox(param, label);
	}
	
	
	public static class ACFreeInput extends ArgComponent{
		
		public final JTextField field;
		private boolean quote;
		
		public ACFreeInput(String param, String label, boolean quote) {
			super(param);
			this.quote = quote;
			setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
			setMaximumSize(new Dimension(Integer.MAX_VALUE, DEFAULT_HEIGHT));
			
			JLabel l = new JLabel(label);
			field = new JTextField();
			
			add(l);
			add(field);
		}

		@Override
		public String getValue() throws InvalidOptionException {
			if(quote)
				return Constants.quote(field.getText());
			return field.getText();
		}

		@Override
		public void setValue(String value) {
			if(quote)
				value = Constants.unquote(value);
			field.setText(value);
		}
	}
	
	public static class ACFileInput extends ACFreeInput{

		private FileFilter[] filters = new FileFilter[0];
		
		public ACFileInput(String param, String label, boolean quote) {
			super(param, label, quote);
			
			JButton ref = new JButton("..."); //$NON-NLS-1$
			ref.addActionListener(new ActionListener(){
				public void actionPerformed(ActionEvent e) {
					File file = FileChooserUtility.chooseOpenFile("ファイル選択", Prefs.WORKING_DIRECTORY, filters, ACFileInput.this); //$NON-NLS-1$
					if(file == null)return;
					File abstractFile = FileUtility.makeAbstractPath(Prefs.WORKING_DIRECTORY, file);
					
					field.setText(abstractFile.getPath());
				}
			});
			
			add(ref);
		}
		
		public void setFileFilter(FileFilter[] filters) {
			this.filters = filters.clone();
		}

		@Override
		public String getValue() throws InvalidOptionException {
			return Constants.replaceFileSeparator(super.getValue());
		}
	}

	public static class ACRadioButtons extends ArgComponent{
		
		public final JRadioButton[] buttons;
		private String[] values;
		public JTextField field;
		private boolean quote;

		public ACRadioButtons(String param, String title, String[] labels, String[] values,
				boolean quote) 
		{
			super(param);
			setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
			setBorder(new TitledBorder(title));
			
			this.values = values.clone();
			this.quote = quote;
			buttons = new JRadioButton[labels.length];
			
			ButtonGroup group = new ButtonGroup();
			for(int i = 0; i < labels.length; i++){
				buttons[i] = new JRadioButton(labels[i]);
				group.add(buttons[i]);
				if(values[i] == null){//自由入力
					if(field != null)
						throw new RuntimeException("cannot make free input > 1");
					Box box = Box.createHorizontalBox();
					field = new JTextField("");
					//field.setHorizontalAlignment(JTextField.RIGHT);
					field.setPreferredSize(new Dimension(50, DEFAULT_HEIGHT));
					//field.setMinimumSize(new Dimension(50, DEFAULT_HEIGHT));
					box.add(buttons[i]);
					box.add(field);
					add(box);
				}else{
					add(buttons[i]);
				}
			}
		}

		@Override
		public String getValue() throws InvalidOptionException {
			for(int i = 0; i < buttons.length; i++){
				if(buttons[i].isSelected()){
					if(values[i] == null){
						if(quote)
							return Constants.quote(field.getText());
						return field.getText();
					}else{
						return values[i];
					}
				}
			}
			throw new RuntimeException();
		}

		@Override
		public void setValue(String value) {
			int freeInput = -1;
			for(int i = 0; i < values.length; i++){
				if(values[i] == null)
					freeInput = i;
				else if(values[i].equals(value)){
					buttons[i].setSelected(true);
					return;
				}
			}
			buttons[freeInput].setSelected(true);
			if(quote)
				value = Constants.unquote(value);
			field.setText(value);
			return;
		}
	}
	
	public static class ACComboBox extends ArgComponent{
		
		public JComboBox comboBox;
		private String[] values;
		
		public ACComboBox(String param, String title, String[] labels, String[] values)
		{
			super(param);
			setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
			setMaximumSize(new Dimension(Integer.MAX_VALUE, DEFAULT_HEIGHT));
			
			this.values = values.clone();
			
			JLabel label = new JLabel(title);
			comboBox = new JComboBox(labels);
			
			add(label);
			add(comboBox);
		}
		
		@Override
		public String getValue() throws InvalidOptionException {
			return values[comboBox.getSelectedIndex()];
		}
		@Override
		public void setValue(String value) {
			for(int i = 0; i < values.length; i++){
				if(values[i].equals(value)){
					comboBox.setSelectedIndex(i);
					return;
				}
			}
		}
	}
	
	public static class ACList extends ArgComponent{

		public JList list;
		protected String[] values;
		private boolean sigleSelection;
		
		public ACList(String param, String title, String[] labels, String[] values,
				boolean sigleSelection) 
		{
			super(param);
			setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
			
			this.values = values.clone();
			this.sigleSelection = sigleSelection;
			
			Box labelBox = Box.createHorizontalBox();
			JLabel label = new JLabel(title);
			labelBox.add(label);
			labelBox.add(Box.createHorizontalGlue());
			list = new JList(labels);
			list.setSelectionMode((sigleSelection)?ListSelectionModel.SINGLE_SELECTION:ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
			
			JScrollPane sp = new JScrollPane(list);
			sp.setPreferredSize(new Dimension(150, 100));
			sp.setMaximumSize(sp.getPreferredSize());
			
			add(labelBox);
			add(sp);
		}
		
		@Override
		public String getValue() throws InvalidOptionException {
			if(sigleSelection)
				return values[list.getSelectedIndex()];
			else{
				int[] selectedIndices = list.getSelectedIndices();
				String[] selected = new String[selectedIndices.length];
				for(int i = 0; i < selected.length; i++){
					selected[i] = values[selectedIndices[i]];
				}
				return Constants.combine(selected);
			}
		}
		@Override
		public void setValue(String value) {
			if(sigleSelection){
				for(int i = 0; i < values.length; i++){
					if(values[i].equals(value)){
						list.setSelectedIndex(i);
						return;
					}
				}
			}else{
				String[] vars = RCommand.getAllVars(value);
				if(vars == null)return;
				HashSet<String> varSet = new HashSet<String>();
				for(String var : vars)varSet.add(var);
				for(int i = 0; i < values.length; i++){
					if(varSet.contains(values[i])){
						list.addSelectionInterval(i, i);
					}
				}
			}
		}
	}
	
	public static class ACVariableChooser extends ACList{

		public JComboBox dataComboBox;
		
		public ACVariableChooser(String param, String title, 
				JComboBox dataComboBox, boolean singleSelection) 
		{
			super(param, title, new String[]{""}, new String[]{""}, singleSelection);
			list.setModel(new DefaultListModel());
			
			this.dataComboBox = dataComboBox;
			
			dataComboBox.addActionListener(new ActionListener(){
				public void actionPerformed(ActionEvent e) {
					updateList();
				}
			});
			
			updateList();
		}
		
		public void updateList(){
			JComboBox comboBox = dataComboBox;
			String dataframe = (String)comboBox.getSelectedItem();
			REXP names = RCommand.getR().eval("names("+dataframe+")");
			String[] c = null;
	        if (names != null && (c=names.asStringArray()) != null){
	        	DefaultListModel listModel = (DefaultListModel)list.getModel();
	        	listModel.clear();
	        	for(String var : c){
	        		listModel.addElement(var);
	        	}
	        	values = c;
	        }	
		}
	}
	
	public static class ACSimpleFormula extends ArgComponent{

		public ACVariableChooser response;
		public ACVariableChooser explain;
		
		public ACSimpleFormula(String param, String title1, String title2,
				JComboBox dataComboBox) 
		{
			super(param);
			
			response = new ACVariableChooser(param, title1, dataComboBox, true);
			explain = new ACVariableChooser(param, title2, dataComboBox, false);
			
			add(response);
			add(explain);
		}
		
		@Override
		public String getValue() throws InvalidOptionException {
			String res = response.getValue();
			String exp = explain.getValue();
			String[] exps = RCommand.getAllVars(exp);
			if(exps == null)throw new InvalidOptionException("");//TODO message
			String expFormula = "";
			for(String e : exps)
				expFormula += e + " + ";
			expFormula = expFormula.substring(0, expFormula.length()-3);
			return res + " ~ " + expFormula;
		}
		
		@Override
		public void setValue(String value) {
			String[] vars = RCommand.getAllVars(value);
			if(vars == null)return;
			response.setValue(vars[0]);
			//TODO .の変換等
			String[] exps = new String[vars.length-1];
			System.arraycopy(vars, 1, exps, 0, exps.length);
			explain.setValue(Constants.combine(exps));
		}
	}

	public static class ACCheckBox extends ArgComponent{
		
		public final JCheckBox checkBox;
		
		public ACCheckBox(String param, String label) {
			super(param);
			setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
			setMaximumSize(new Dimension(Integer.MAX_VALUE, DEFAULT_HEIGHT));
			
			checkBox = new JCheckBox(label);
			
			add(checkBox);
			add(getOperativeLabel());
		}

		@Override
		public String getValue() throws InvalidOptionException {
			return Constants.toRBoolean(checkBox.isSelected());
		}

		@Override
		public void setValue(String value) {
			try{
				checkBox.setSelected(Constants.toBoolean(value));
				setOperative(true);
			}catch(IllegalArgumentException e){
				setOperative(false);
			}
		}
	}


}
