package com.ef_prime.rflow.node;

import java.awt.BorderLayout;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JOptionPane;
import javax.swing.JTabbedPane;
import javax.swing.border.EmptyBorder;

import com.ef_prime.rflow.exception.CanceledException;
import com.ef_prime.rflow.exception.InvalidOptionException;
import com.ef_prime.rflow.resource.Messages;
import com.ef_prime.rflow.tool.ButtonPanel;


public abstract class DefaultNodeModelController extends AbstractNodeModelController{
	
	private NodeModel model = null;
	
	private Window parent;
	
	public NodeModel createNewModel() throws CanceledException {
		return getNodeModel(getDefaultModelInstance());
	}
	
	public Window getParentWindow() {
		return parent;
	}
	
	public void setParentWindow(Window parent) {
		this.parent = parent;
	}
	
	/**
	 * 初期値が設定されたモデルの新しいインスタンスを返すメソッドです。
	 * 引数なしのコンストラクタを使ってインスタンスを作成します。
	 * @return
	 */
	protected NodeModel getDefaultModelInstance(){
		try {
			return getModelClass().newInstance();
		} catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("Failed to create default model Instance."); //$NON-NLS-1$
		}
	}

	protected abstract NodeOptionPanel createNodeOptionPanel(NodeModel model);
	
	protected final void requestOkButtonFocus(){
		if(ok != null && ok.isVisible())
			ok.requestFocus();
	}
	
	public NodeModel getNodeModel(NodeModel orgModel) throws CanceledException{
		model = null;
		NodeOptionDialog dialog;
		Window window = getParentWindow();
		if (window instanceof Frame) 
			dialog= new NodeOptionDialog((Frame)window, orgModel.copy());
		else
			dialog= new NodeOptionDialog((Dialog)window, orgModel.copy());
		dialog.setVisible(true);
		dialog.dispose();
		if(model == null){
			throw new CanceledException();
		}
		return model;
	}
	
	
	private JButton ok;//OK Button only for requestFocus
	private class NodeOptionDialog extends JDialog{

		NodeOptionPanel panel;
		NodePropertyPanel propertyPanel;
		
		public NodeOptionDialog(Dialog parent, NodeModel model){
			super(parent, Messages.getString("DefaultNodeModelController.title"), true); //$NON-NLS-1$
			init(parent, model);
		}
		
		public NodeOptionDialog(Frame parent, NodeModel model){
			//super(parent, "編集", ModalityType.APPLICATION_MODAL);
			super(parent, Messages.getString("DefaultNodeModelController.title"), true); //$NON-NLS-1$
			
			init(parent, model);
		}

		private void init(Window parent, NodeModel model) {
			panel = createNodeOptionPanel(model);
			propertyPanel = new NodePropertyPanel(model);
			
			JTabbedPane tabbedPane = new JTabbedPane();
			tabbedPane.add(panel, Messages.getString("DefaultNodeModelController.setting")); //$NON-NLS-1$
			tabbedPane.add(propertyPanel, Messages.getString("DefaultNodeModelController.property")); //$NON-NLS-1$
			
			
			ActionListener okAction = new ActionListener(){
				public void actionPerformed(ActionEvent e) {
					
					try {
						NodeOptionDialog.this.setOption();
					} catch (InvalidOptionException e1) {
						JOptionPane.showMessageDialog(getParentWindow(), e1.getMessage(), Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$
						return;
					}
					NodeOptionDialog.this.dispose();
				}
			};
			ActionListener cancelAction = new ActionListener(){
				public void actionPerformed(ActionEvent e) {
					DefaultNodeModelController.this.model = null;
					NodeOptionDialog.this.dispose();
				}
			};
			
			ButtonPanel buttonPanel = new ButtonPanel(okAction, cancelAction);
			Box btnBox = Box.createHorizontalBox();
			btnBox.add(Box.createHorizontalGlue());
			btnBox.add(buttonPanel);
			btnBox.add(Box.createHorizontalGlue());
			btnBox.setMaximumSize(new Dimension(Integer.MAX_VALUE, 100));
			ok = buttonPanel.getOkButton();
			
			Box box = Box.createVerticalBox();
			int sp = 20;
			box.setBorder(new EmptyBorder(sp,sp,sp,sp));
			box.add(tabbedPane);
			box.add(Box.createVerticalStrut(20));
			box.add(btnBox);
			
			getContentPane().setLayout(new BorderLayout());
			getContentPane().add(box);
			
			pack();
			setLocationRelativeTo(parent);
			
			panel.initFocus();
		}
		
		public void setOption() throws InvalidOptionException{
			model = panel.getModel();
			propertyPanel.setPropertiesToModel(model);
		}
		
		/*
		private class TitlePanel extends JPanel implements OptionSetter{

			JTextField titleText;
			
			TitlePanel(){
				super();
				setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
				
				JLabel label = new JLabel("タイトル");
				label.setMaximumSize(labelSize);
				label.setPreferredSize(labelSize);
				
				titleText = new JTextField();
				titleText.setText(model.title);
				
				
				add(label);
				add(titleText);
				
			}
			
			public void setOption() {
				model.title = titleText.getText();
			}
		}*/
		
		/*private class ButtonPanel extends JPanel{
			ButtonPanel(){
				super();
				setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
				
				
				ok = createButton("OK");
				ok.addActionListener(new ActionListener(){
					public void actionPerformed(ActionEvent e) {
						NodeOptionDialog.this.setOption();
						NodeOptionDialog.this.dispose();
					}
				});
				JButton cancel = createButton("Cancel");
				cancel.addActionListener(new ActionListener(){
					public void actionPerformed(ActionEvent e) {
						model = null;
						NodeOptionDialog.this.dispose();
					}
				});
				
				add(ok);
				add(Box.createHorizontalStrut(50));
				add(cancel);
			}
			
			JButton createButton(String title){
				JButton button = new JButton(title);
				Dimension buttonSize = new Dimension(80, 25);
				button.setPreferredSize(buttonSize);
				button.setMaximumSize(buttonSize);
				
				return button;
			}
		}*/
	}
	
}
