package com.ef_prime.rflow.node;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import com.ef_prime.rflow.RCommand;
import com.ef_prime.rflow.Rflow;
import com.ef_prime.rflow.exception.IncompatibleCommandException;
import com.ef_prime.rflow.node.base.ArgEntry;
import com.ef_prime.rflow.node.base.ArgList;
import com.ef_prime.rflow.node.base.SentenceNodeModel;

public abstract class SimpleNodeModel extends AbstractNodeModel implements Cloneable{

	protected ArgList argList;
	private String symbol = "";
	private String function = "";
	
	public SimpleNodeModel(){
		super();
		//function = getDefaultFunction();
		//argList = new ArgList(getDefaultArgEntryList());
	}
	
	protected abstract String getDefaultFunction();
	
	protected List<ArgEntry> getDefaultArgEntryList(){
		Map<String, String> map = RCommand.argList(function);
		List<ArgEntry> list =  new ArrayList<ArgEntry>();
		if(map == null)
			return list;
		for(Map.Entry<String, String> entry : map.entrySet())
			list.add(new ArgEntry(entry.getKey(), entry.getValue()));
		return list;
	}
	
	protected static List<ArgEntry> getArgEntryList(String[] array){
		List<ArgEntry> list = new ArrayList<ArgEntry>();
		for(int i = 0; i < array.length; i += 2){
			list.add(new ArgEntry(array[i], array[i+1]));
		}
		return list;
	}
	
	public ArgList getArglist() {
		return argList;
	}
	
	public String getValue(String param){
		return argList.getValue(param);
	}

	public String putArgument(String param, String value) {
		return argList.putArgument(param, value);
	}

	public String getCommand() {
		String assign = (symbol.length()>0)?symbol+" <- ":"";
		return assign + function + argList.list(false);
	}

	public String getFunction() {
		return function;
	}
	
	public String getSymbol() {
		return symbol;
	}

	public void setFunction(String function) {
		if(function == null)
			throw new IllegalArgumentException("function cannot be null");
		this.function = function;
	}
	
	public void setSymbol(String symbol) {
		if(symbol == null)
			throw new IllegalArgumentException("symbol cannot be null");
		this.symbol = symbol;
	}
	
	@Override
	public boolean isEditable() {
		return true;
	}
	
	private static final int MAX_TRY = 1;
	@Override
	public void setCommand(String command)throws IncompatibleCommandException{
		if(command.equals(getCommand()))
			return;
		/*
		 * MAX_TRY回コマンド
		 */
		int count = MAX_TRY;
		while(count > 0 && !extractFunction(command)){
			count--;
			if(Rflow.DEBUG > 0)
				System.out.println("IncompatibleCommandException:"+(MAX_TRY-count)); //$NON-NLS-1$
			if(count == 0)
				throw new IncompatibleCommandException();
		}
	}
	
	private boolean extractFunction(String command){
		//TODO SentenceNodeModelをかいする必要なし
		NodeModel[] models = RCommand.parseToNodes(command, false);
		if(models == null || models.length > 1){
			return false;
		}
		if(!(models[0] instanceof SentenceNodeModel))
			return false;
		
		SentenceNodeModel model = (SentenceNodeModel)models[0];
		symbol = (model.symbol == null)?"":model.symbol;
		function = (model.function == null)?"":model.function;
		argList.clear();
		for(ArgEntry entry : model.argList){
			argList.putArgument(entry.param, entry.value);
		}
		
		return true;
	}

	@Override
	public NodeModel copy() {
		try {
			return (NodeModel)clone();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
			throw new RuntimeException("CloneNotSupportedException");
		}
	}
}
