package com.ef_prime.rflow.node.base;

import java.awt.Shape;
import java.util.ArrayList;
import java.util.List;

import javax.swing.Icon;

import com.ef_prime.rflow.Constants;
import com.ef_prime.rflow.RCommand;
import com.ef_prime.rflow.Rflow;
import com.ef_prime.rflow.exception.IncompatibleCommandException;
import com.ef_prime.rflow.flow.rendering.NodeShape;
import com.ef_prime.rflow.group.AppearanceGroup;
import com.ef_prime.rflow.group.AppearanceGroupManager;
import com.ef_prime.rflow.node.AbstractNodeModel;
import com.ef_prime.rflow.node.NodeModel;

public class SentenceNodeModel extends AbstractNodeModel{

	public String sentence = ""; //$NON-NLS-1$
	
	public String symbol = null;

	public String function = null;
	public List<ArgEntry> argList = new ArrayList<ArgEntry>();
	
	private AppearanceGroup group;
	
	
	public SentenceNodeModel() {
		super();
	}
	
	private SentenceNodeModel(SentenceNodeModel model) {
		super(model);
		sentence = model.sentence;
		if(model.isAssign())
			symbol = model.symbol;
		if(model.isFunction()){
			function = model.function;
			argList = new ArrayList<ArgEntry>(model.argList);
		}
		group = model.group;
	}

	@Override
	public boolean isEditable() {
		return !getCommand().contains(Constants.NEWLINE_STR);
	}
	

	private static final int MAX_TRY = 1;
	public void setCommand(String s)throws IncompatibleCommandException{
		if(s.equals(getCommand()))
			return;
		/*
		 * MAX_TRY回コマンド
		 */
		int count = MAX_TRY;
		while(count > 0 && !extractFunction(s)){
			count--;
			if(Rflow.DEBUG > 0)
				System.out.println("IncompatibleCommandException:"+(MAX_TRY-count)); //$NON-NLS-1$
			if(count == 0)
				throw new IncompatibleCommandException();
		}
	}
	
	private boolean extractFunction(String command){
		NodeModel[] models = RCommand.parseToNodes(command, false);
		if(models == null || models.length > 1){
			return false;
		}
		if(!(models[0] instanceof SentenceNodeModel))
			return false;
		
		SentenceNodeModel model = (SentenceNodeModel)models[0];
		//sentence = model.sentence;
		sentence = ""; //$NON-NLS-1$
		for(String s : model.sentence.split("\n")) //$NON-NLS-1$
			sentence += s.trim();
		symbol = model.symbol;
		function = model.function;
		for(ArgEntry entry : model.argList){
			String value = ""; //$NON-NLS-1$
			for(String s : entry.value.split("\n")) //$NON-NLS-1$
				value += s.trim();
			entry.value = value;
		}
		argList = model.argList;
		
		return true;
	}
	
	public String getCommandFromFunction(){
		if(!isFunction())return null;
		String command = ""; //$NON-NLS-1$
		if(isAssign()){
			command += symbol + " <- ";  //$NON-NLS-1$
		}
		command += function;
		
		command += "("; //$NON-NLS-1$
		int argCount = 0;
		for(ArgEntry entry : argList){
			if(!entry.isValue())continue;
			argCount++;
			if(entry.isParam() && entry.isValue()){
				command += entry.param + "="; //$NON-NLS-1$
			}
			if(entry.isValue()){
				command += entry.value + ", "; //$NON-NLS-1$
			}
		}
		if(argCount > 0)
			command = command.substring(0, command.length()-2);
		command += ")"; //$NON-NLS-1$
		return command;
	}

	
	public boolean isFunction(){
		return function != null;
	}

	public boolean isAssign(){
		return (symbol != null) && (symbol.length() > 0);
	}
	
	@Override
	protected String getDefaultTitle() {
		if(isFunction()){
			String title = ""; //$NON-NLS-1$
			if(isAssign()){
				title += symbol + "<-"; //$NON-NLS-1$
			}
			title += function;
			return title;
		}else{
			String c = getCommand();
			if(c.length() > 10)
				return getCommand().substring(0, 10)+"..."; //$NON-NLS-1$
			else
				return c;
		}
	}
	
	@Override
	public NodeModel copy() {		
		return new SentenceNodeModel(this);
	}

	public String getCommand() {
		String command = ""; //$NON-NLS-1$
		if(isAssign())
			command += symbol + " <- ";  //$NON-NLS-1$
		command += sentence;
		return command;
	}

	//-----------------------------------------------------------------
	//shape
	//-----------------------------------------------------------------
	
	@Override
	public Icon getIcon() {
		if(group == null)
			return null;
		return group.getIcon();
	}
	
	@Override
	public Icon getSmallIcon() {
		if(group == null)
			return super.getSmallIcon();
		return group.getSmallIcon();
	}
	
	@Override
	public Shape getShape() {
		if(group != null)
			return group.getShape();
		
		return getNodeShape().getShape();
	}
	
	@Override
	protected NodeShape getNodeShape() {
		if(isAssign())
			return NodeShape.CIRCLE;
		else
			return NodeShape.TRIANGLE;
	}
	
	@Override
	public Shape getGlassShape() {
		if(group != null)
			return group.getGlassShape();
		return super.getGlassShape();
	}
	
	@Override
	public void update(AppearanceGroupManager groupManager) {
		group = groupManager.getAppearanceGroup(function);
	}
	
	public void setGroup(AppearanceGroup group){
		this.group = group;
	}

	@Override
	public int hashCode() {
		return super.hashCode();
	}
	
	@Override
	public boolean equals(Object obj) {
		if (obj instanceof SentenceNodeModel) {
			return super.equals(obj);
		}
		return false;		
	}
}
