package com.ef_prime.rflow.node.base;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.Map;
import java.util.Properties;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;
import javax.swing.border.LineBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.CellEditorListener;
import javax.swing.event.ChangeEvent;
import javax.swing.event.DocumentEvent;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;

import com.ef_prime.rflow.Prefs;
import com.ef_prime.rflow.RCommand;
import com.ef_prime.rflow.exception.IncompatibleCommandException;
import com.ef_prime.rflow.exception.ModelCreationFailedException;
import com.ef_prime.rflow.node.DefaultNodeModelController;
import com.ef_prime.rflow.node.NodeModel;
import com.ef_prime.rflow.node.NodeOptionPanel;
import com.ef_prime.rflow.resource.Messages;
import com.ef_prime.rflow.tool.CodingTextPane;
import com.ef_prime.rflow.tool.SupportControl;
import com.ef_prime.rflow.tool.SupportedInputFactory;
import com.ef_prime.rflow.util.DocumentAdapter;

public class SentenceNodeModelController extends DefaultNodeModelController{

	public String getModelName() {
		return Messages.getString("SentenceNodeModelController.model_name"); //$NON-NLS-1$
	}
	
	public Class<? extends NodeModel> getModelClass() {
		return SentenceNodeModel.class;
	}
	
	//@Override
	/*public boolean checkModelClass(NodeModel model) {
		return model instanceof SentenceNodeModel;
	}*/
	
	public Properties getOptionProperties(NodeModel model) {
		return null;
	}
	
	public NodeModel createModel(Properties properties, String title, String command) throws ModelCreationFailedException {
		SentenceNodeModel model = new SentenceNodeModel();
		try {
			model.setCommand(command);
		} catch (IncompatibleCommandException e) {
			throw new ModelCreationFailedException();
		}
		return model;
	}	

	@Override
	protected NodeOptionPanel createNodeOptionPanel(NodeModel model) {
		return new SentenceNodeOptionPanel(model);
	}

	private class SentenceNodeOptionPanel extends NodeOptionPanel{

		FunctionCodePanel functionCodePanel;
		AssignOptionPanel assignOption;
		FunctionOptionPanel functionOption;
		
		SentenceNodeModel model;
		
		public SentenceNodeOptionPanel(NodeModel model){
			super();
			this.model = (SentenceNodeModel)model;
			
			functionCodePanel = new FunctionCodePanel();
			assignOption = new AssignOptionPanel();
			functionOption = new FunctionOptionPanel();

			int space = 20;
			
			Box box = Box.createVerticalBox();
			box.setBorder(new EmptyBorder(space,space,space,space));
			box.add(functionCodePanel);
			//box.add(Box.createVerticalStrut(20));
			box.add(assignOption);
			//box.add(Box.createVerticalStrut(20));
			box.add(functionOption);
			
			//setFocusTraversalPolicy(FocusTraversalPolicy);
			this.setLayout(new BorderLayout());
			this.add(box, BorderLayout.CENTER);
			box.addMouseListener(new MouseAdapter(){
				@Override
				public void mousePressed(MouseEvent e) {
					requestFocus();
				}
			});
			
			
			if(functionCodePanel.textPane.getText().length() > 0)
				functionCodePanel.editFinished();
		}
		
		@Override
		protected void initFocus() {
			functionCodePanel.textPane.requestFocus();
		}
		
		private void setFunctionOptionEnable(boolean b){
			//setEnable(assignOption, b);
			setEnable(functionOption, b);
		}
		
		private void setEnable(Container c, boolean b){
			for(Component child : c.getComponents()){
				if(child instanceof Container){
					child.setEnabled(b);
					setEnable((Container)child, b);
				}else{
					child.setEnabled(b);
				}
			}
		}
		
		@Override
		protected NodeModel getModel() {
			//String command = functionCodePanel.textPane.getText();
			//SentenceNodeModel model = new SentenceNodeModel();
			//model.setCommand(command);
			return model;
		}		

		private boolean lockCode2Options = false;
		private void code2Options(SentenceNodeModel model) {
			if(lockCode2Options)return;

			lockOption2Code = true;
			assignOption.evalOnlyButton.setSelected(!model.isAssign());
			assignOption.assignButton.setSelected(model.isAssign());
			if(model.isAssign())
				assignOption.assignTarget.setText(model.symbol);
			//else
			//	assignOption.assignTarget.setText("");
			
			if(model.isFunction()){
				functionOption.control.setAutoSupport(false);
				functionOption.functionText.setText(model.function);
				functionOption.control.setAutoSupport(true);
				functionOption.table.setArgsList(model.argList);
				
				functionOption.argsComplete(model.function);
			}
			lockOption2Code = false;
			
			//functionCodePanel.setText(fModel.getCommand());
		}
		
		private KeyListener keylistener = new KeyAdapter(){
			@Override
			public void keyPressed(KeyEvent e) {
				if(e.isConsumed())return;
				Component c = (Component)e.getSource();
				switch(e.getKeyCode()){
				case KeyEvent.VK_ENTER:
					c.transferFocus();
					e.consume();
					break;
				}
			}
		};
		
		private boolean lockOption2Code = false;
		private void option2Code(){
			if(!lockOption2Code){
				lockCode2Options = true;
				assignOption.updateModel(model);
				if(model.isFunction()){
					functionOption.updateModel(model);
					functionCodePanel.setText(model.getCommandFromFunction());
				}else{
					functionCodePanel.setText(model.getCommand());
				}
				lockCode2Options = false;
			}
		}
		
		private class FunctionCodePanel extends JPanel implements FocusListener{
			/*
			private Color editing_background = Color.WHITE;
			private Color non_editing_background = new Color(240,240,240); 
			*/
			private JLabel msgLabel;
			private CodingTextPane textPane;
			
			protected FunctionCodePanel(){
				super();
				setLayout(new BorderLayout());
				setBorder(new TitledBorder(Messages.getString("SentenceNodeModelController.code"))); //$NON-NLS-1$
				
				msgLabel = new JLabel(" "); //$NON-NLS-1$
				
				final JPanel borderPanel = new JPanel(new BorderLayout());
				
				textPane = new CodingTextPane();
				textPane.setText(model.getCommand());
				textPane.setParenthesisSupport(true);
				textPane.addFocusListener(this);
				//textPane.addKeyListener(keylistener);
				textPane.addKeyListener(new KeyAdapter(){
					@Override
					public void keyPressed(KeyEvent e) {
						if(e.isConsumed())return;
						switch(e.getKeyCode()){
						case KeyEvent.VK_ENTER:
							requestOkButtonFocus();
							e.consume();
							break;
						}
					}
				});
				textPane.getDocument().addDocumentListener(new DocumentAdapter(){
					@Override
					public void documentUpdate(DocumentEvent e) {
						if(e.getDocument().getLength() == 0)
							return;
						editFinished();
					}
				});
				final Border focusBorder = new LineBorder(Color.GRAY, 1); 
				//final Border focusBorder = new LineBorder(new Color(0,0.5f,1.0f,0.3f), 1); 
				final Border noFocusBorder = new LineBorder(Color.WHITE, 1);
				borderPanel.setBorder(noFocusBorder);
				textPane.addFocusListener(new FocusListener(){
					public void focusGained(FocusEvent e) {
						borderPanel.setBorder(focusBorder);
					}

					public void focusLost(FocusEvent e) {
						borderPanel.setBorder(noFocusBorder);
					}					
				});
				
				borderPanel.add(textPane);
				
				add(borderPanel);
				add(msgLabel, BorderLayout.NORTH);
			}
			
			public void setText(String s){
				textPane.setText(s);
			}
			
			public boolean editFinished(){
				String c = textPane.getRawText();
				try {
					model.setCommand(c);
					
					code2Options(model);
					if(model.isFunction()){
						msgLabel.setText(" "); //$NON-NLS-1$
						setFunctionOptionEnable(true);
					}else{
						msgLabel.setText(Messages.getString("SentenceNodeModelController.extract_function_failed")); //$NON-NLS-1$
						setFunctionOptionEnable(false);
					}
					//textPane.setBackground(non_editing_background);
				} catch (IncompatibleCommandException e) {
					//textPane.setBackground(editing_background);
					msgLabel.setText(Messages.getString("SentenceNodeModelController.syntax_error_or_multiline")); //$NON-NLS-1$
					setFunctionOptionEnable(false);
					//textPane.setText(getModel().getCommand());
				}
				
				return true;
			}

			public void focusGained(FocusEvent e) {
				//textPane.setBackground(editing_background);
			}

			public void focusLost(FocusEvent e) {
				editFinished();
			}
		}
		
		private class AssignOptionPanel extends JPanel{
			
			JRadioButton evalOnlyButton;
			JRadioButton assignButton;
			JTextField assignTarget;
			
			protected AssignOptionPanel(){
				super();
				setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
				setBorder(new TitledBorder(Messages.getString("SentenceNodeModelController.assign_option"))); //$NON-NLS-1$
				
				evalOnlyButton = new JRadioButton(Messages.getString("SentenceNodeModelController.eval_only"), !model.isAssign()); //$NON-NLS-1$
				assignButton = new JRadioButton(Messages.getString("SentenceNodeModelController.assign"), model.isAssign()); //$NON-NLS-1$
				assignButton.addItemListener(new ItemListener(){
					public void itemStateChanged(ItemEvent e) {
						assignTarget.setEnabled(assignButton.isSelected());
						option2Code();
					}
				});
				assignTarget = new JTextField();
				assignTarget.setFont(Prefs.CODE_FONT_12);
				assignTarget.setEnabled(assignButton.isSelected());
				assignTarget.setMaximumSize(HEIGHTLIMIT_LABEL_SIZE);
				assignTarget.addKeyListener(keylistener);
				assignTarget.addFocusListener(new FocusAdapter(){
					@Override
					public void focusLost(FocusEvent e) {
						option2Code();
					}
				});
				if(model.symbol != null)assignTarget.setText(model.symbol);
				
				ButtonGroup group = new ButtonGroup();

				group.add(evalOnlyButton);
				group.add(assignButton);
				
				Box evalBox = Box.createHorizontalBox();
				evalBox.add(evalOnlyButton);
				evalBox.add(Box.createHorizontalGlue());
				
				Box symbolBox = Box.createHorizontalBox();
				symbolBox.add(assignButton);
				symbolBox.add(Box.createHorizontalStrut(30));
				symbolBox.add(assignTarget);
				
				add(evalBox);
				add(symbolBox);
			}
			
			public void updateModel(SentenceNodeModel model){
				if(assignButton.isSelected())
					model.symbol = assignTarget.getText();
				else
					model.symbol = null;
			}
		}
		
		private class FunctionOptionPanel extends JPanel {
			
			
			JTextField functionText;
			SupportControl control;
			ArgsTable table;
			
			protected FunctionOptionPanel(){
				super();
				setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));	
				setBorder(new TitledBorder(Messages.getString("SentenceNodeModelController.function_option"))); //$NON-NLS-1$
				
				JLabel functionLabel = new JLabel(Messages.getString("SentenceNodeModelController.function")); //$NON-NLS-1$
				functionLabel.setMaximumSize(DEFAULT_LABEL_SIZE);
				functionLabel.setPreferredSize(DEFAULT_LABEL_SIZE);
				functionText = new JTextField();
				functionText.setFont(Prefs.CODE_FONT_12);
				functionText.setMaximumSize(HEIGHTLIMIT_LABEL_SIZE);
				if(model.isFunction())functionText.setText(model.function);
				functionText.addFocusListener(new FocusAdapter(){
					@Override
					public void focusLost(FocusEvent e) {
						option2Code();
					}
				});
				functionText.addKeyListener(new KeyAdapter(){
					public void keyPressed(KeyEvent e) {
						if(control.isSupportPopup())return;
						if(e.isConsumed())return;
						Component c = (Component)e.getSource();
						switch(e.getKeyCode()){
						case KeyEvent.VK_ENTER:
							c.transferFocus();
							argsComplete(functionText.getText());
							e.consume();
							break;
						}
					};
				});
				
				control = SupportedInputFactory.setFunctionSupport(functionText);
				control.setAutoSupport(true);
				/*control.addItemAcceptListener(new ItemAcceptListener(){
					public void itemAccepted(JTextComponent source, String accepted) {
						argsComplete(accepted);
					}
				});*/
				functionText.getDocument().addDocumentListener(new DocumentAdapter(){
					@Override
					public void documentUpdate(DocumentEvent e) {
						try {
							Document document = e.getDocument();
							String function = document.getText(0, document.getLength());
							table.setFunction(function);
							argsComplete(function);
						} catch (BadLocationException e1) {e1.printStackTrace();}
					}
				});
				
				//Dimension littelButtonSize = new Dimension(45,20);
				
				JButton sourceButton = new JButton(Messages.getString("SentenceNodeModelController.src")); //$NON-NLS-1$
				//sourceButton.setMaximumSize(littelButtonSize);
				//sourceButton.setPreferredSize(littelButtonSize);
				sourceButton.addActionListener(new ActionListener(){
					public void actionPerformed(ActionEvent e) {
						RCommand.getReceiver().passCommand(functionText.getText());
					}
				});
				
				JButton helpButton = new JButton(Messages.getString("SentenceNodeModelController.help")); //$NON-NLS-1$
				//helpButton.setMaximumSize(littelButtonSize);
				//helpButton.setPreferredSize(littelButtonSize);
				helpButton.addActionListener(new ActionListener(){
					public void actionPerformed(ActionEvent e) {
						RCommand.getReceiver().passCommand("help("+functionText.getText()+")"); //$NON-NLS-1$ //$NON-NLS-2$
					}
				});
				
				Box functionBox = Box.createHorizontalBox();
				functionBox.add(functionLabel);
				functionBox.add(functionText);
				functionBox.add(sourceButton);
				functionBox.add(helpButton);
				
				
				JLabel argsLabel = new JLabel(Messages.getString("SentenceNodeModelController.parameter")); //$NON-NLS-1$
				argsLabel.setMaximumSize(DEFAULT_LABEL_SIZE);
				argsLabel.setPreferredSize(DEFAULT_LABEL_SIZE);
				table = new ArgsTable();
				if(model.argList != null)
					table.setArgsList(model.argList);
				table.getTableCellEditor().addCellEditorListener(new CellEditorListener(){
					public void editingCanceled(ChangeEvent e) {}
					public void editingStopped(ChangeEvent e) {
						option2Code();
					}
				});
				table.getModel().addTableModelListener(new TableModelListener(){
					public void tableChanged(TableModelEvent e) {
						option2Code();
					}
				});
				JScrollPane pane = new JScrollPane(table);
				pane.setPreferredSize(new Dimension(300, 200));
				
				Box argsBox = Box.createHorizontalBox();
				argsBox.add(argsLabel);
				argsBox.add(pane);
				
				JButton argsSupportButton = new JButton(Messages.getString("SentenceNodeModelController.parameter_complete")); //$NON-NLS-1$
				//argsSupportButton.setMaximumSize(DEFAULT_LABEL_SIZE);
				//argsSupportButton.setPreferredSize(DEFAULT_LABEL_SIZE);
				argsSupportButton.addActionListener(new ActionListener(){
					public void actionPerformed(ActionEvent e) {
						String function = functionText.getText();
						argsComplete(function);
					}
				});
				
				Box argsSupportBox = Box.createHorizontalBox();
				argsSupportBox.setFocusable(true);
				argsSupportBox.add(Box.createHorizontalGlue());
				argsSupportBox.add(argsSupportButton);
				
				
				add(functionBox);
				add(argsBox);
				//add(argsSupportBox);//引数補完ボタン
				
				//
				table.setFunction(functionText.getText());
			}

			private void argsComplete(String function){
				Map<String, String> map = RCommand.argList(function);
				if(map == null || map.size() < 1){
					table.clearDefaultMap();
					return;
				}
				
				table.clearDefaultMap();
				Map<String, String> defaultMap = table.getDefaultMap();
				for(Map.Entry<String, String> entry : map.entrySet()){
					defaultMap.put(entry.getKey(), entry.getValue());
					table.addRow(new ArgEntry(entry.getKey(), entry.getValue()));
				}
				
				//table.setDefaultMap(defaultMap);
			}
			
			public void updateModel(SentenceNodeModel model) {
				model.function = functionText.getText();
				model.argList = table.getArgsList();
			}
		}
		
	}
}
