package com.ef_prime.rflow.node.base;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Window;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JDialog;
import javax.swing.event.AncestorEvent;
import javax.swing.event.AncestorListener;

import com.ef_prime.rflow.Rflow;
import com.ef_prime.rflow.exception.CanceledException;
import com.ef_prime.rflow.flow.FlowUtil;
import com.ef_prime.rflow.flow.FlowViewer;
import com.ef_prime.rflow.flow.ViewLayout;
import com.ef_prime.rflow.flow.model.FlowEdge;
import com.ef_prime.rflow.flow.model.FlowGraph;
import com.ef_prime.rflow.flow.model.FlowNode;
import com.ef_prime.rflow.node.AbstractNodeModel;
import com.ef_prime.rflow.node.NodeModel;

import edu.uci.ics.jung.graph.Vertex;

public class SubflowNodeModel extends AbstractNodeModel{
	
	private ViewLayout layout;
	//private FlowNode in;
	private FlowNode out;
	
	private JDialog contentsDialog;
	private FlowViewer smallViewer;
	
	public SubflowNodeModel(){
		out = new FlowNode(TunnelNodeModel.outModel);
		
		FlowGraph graph = new FlowGraph();
		graph.addVertex(out);
		
		Map<Vertex, Point2D> map = new HashMap<Vertex, Point2D>();
		map.put(out, new Point2D.Double(250, 100));
		
		
		layout = new ViewLayout(graph, map, new Dimension(300, 200), new Point(0, 0));
	}
	
	public SubflowNodeModel(ViewLayout layout, FlowNode in, FlowNode out) {
		this.layout = layout;
		
		//this.in = (FlowNode)layout.getGraph().addVertex(new FlowNode(TunnelNodeModel.getInTunnelModel()));
		this.out = (FlowNode)layout.getGraph().addVertex(new FlowNode(TunnelNodeModel.outModel));
		
		//layout.getGraph().addEdge(new FlowEdge(this.in, in));
		layout.getGraph().addEdge(new FlowEdge(out, this.out));
		
		int offset = 50;
		//Point2D inPoint = layout.getLocationMap().get(in);
		Point2D outPoint = layout.getLocationMap().get(out);
		//layout.getLocationMap().put(this.in, new Point2D.Double(inPoint.getX()-offset, inPoint.getY()));
		layout.getLocationMap().put(this.out, new Point2D.Double(outPoint.getX()+offset, outPoint.getY()));
	}
	
	public SubflowNodeModel(ViewLayout layout){
		this.layout = layout;
		
		for(Object o : layout.getGraph().getVertices()){
			FlowNode node = (FlowNode)o;
			if(node.getModel().equals(TunnelNodeModel.outModel)){
				out = node;break;
			}
		}
	}
	
	private SubflowNodeModel(SubflowNodeModel model){
		super(model);
		this.layout = model.layout;
		
		for(Object o : layout.getGraph().getVertices()){
			FlowNode node = (FlowNode)o;
			if(node.getModel().equals(TunnelNodeModel.outModel)){
				out = node;break;
			}
		}
	}
	
	@Override
	protected String getDefaultTitle() {
		return "BOX"; //$NON-NLS-1$
	}

	public String getCommand() {
		if(out.inDegree() != 1)
			return ""; //$NON-NLS-1$
		FlowNode node = (FlowNode)out.getPredecessors().iterator().next();
		return FlowUtil.getCommandFromStream(FlowUtil.getStream(node), false);
	}
	
	@Override
	public NodeModel showConfigDialog(Window parent) throws CanceledException {
		throw new CanceledException();
	}
	
	public void showContentsDialog(FlowViewer parentViewer, Rflow rflow, Point p){
		if(contentsDialog != null && contentsDialog.isVisible()){
			contentsDialog.requestFocus();
			return;
		}
		

		smallViewer = new FlowViewer(rflow, parentViewer.getPickedState());
		smallViewer.setPreferredSize(layout.getSize());
		smallViewer.init(layout);
		/*smallViewer.addComponentListener(new ComponentAdapter(){
			@Override
			public void componentResized(ComponentEvent e) {
				FlowViewer viewer = (FlowViewer)e.getSource();
				viewer.getGraphLayout().resize(viewer.getSize());
			}
		});*/
		smallViewer.addAncestorListener(new AncestorListener(){
			public void ancestorAdded(AncestorEvent event) {}
			public void ancestorMoved(AncestorEvent event) {}
			public void ancestorRemoved(AncestorEvent event) {
				layout = smallViewer.getViewLayout();
				smallViewer = null;
			}			
		});
		
		contentsDialog = new JDialog(rflow.frame, getTitle());
		contentsDialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
		contentsDialog.getContentPane().add(smallViewer);
		contentsDialog.pack();
		contentsDialog.setLocation(p.x - contentsDialog.getWidth()/2, p.y + 50);
		contentsDialog.setVisible(true);
	}

	@Override
	public NodeModel copy() {
		//TODO copy not supported
		//layoutはコピーされない
		SubflowNodeModel model = new SubflowNodeModel(this);
		return model;
	}

	public ViewLayout getLayout() {
		update();
		return layout;
	}
	
	public void update(){
		if(smallViewer != null && smallViewer.isVisible())
			layout = smallViewer.getViewLayout();
	}

	@Override
	public int hashCode() {
		return super.hashCode();
	}
	
	@Override
	public boolean equals(Object obj) {
		//TODO BoxNodeModel equals実装
		/*if (obj instanceof BoxNodeModel) {
			BoxNodeModel model = (BoxNodeModel) obj;

			return 
				super.equals(model)
				&&
				ObjectUtilities.equals(layout, model.layout);
		}*/
		return false;
	}
}
