package com.ef_prime.rflow.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.channels.FileChannel;

public class FileUtility {
	/**
	 * ファイルをコピーします
	 * @param from
	 * @param to
	 * @throws IOException
	 */
	public static void copy(File from, File to) throws IOException{
		FileInputStream inStream = null;
		FileOutputStream outStream = null;
		FileChannel in = null;
		FileChannel out = null;
		
		try{
			inStream = new FileInputStream(from);
			outStream = new FileOutputStream(to);
			in = inStream.getChannel();
			out = outStream.getChannel();
	
	        in.transferTo(0, in.size(), out);
	        
		}finally{
			if(inStream != null)
				inStream.close();
			if(outStream != null)
				outStream.close();
			if(in != null)
				in.close();
			if(out != null)
				out.close();
		}
	}
	
	/**
	 * absoluteのパスをparentからの抽象パスに変換します
	 * @param parent
	 * @param absolute
	 * @return
	 */
	public static File makeAbstractPath(File parent, File absolute){
		if(parent == null)return absolute;
		
		String parentPath = parent.getAbsolutePath();
		String absPath = absolute.getAbsolutePath();
		File abstractFile = absolute;
		if(isChild(parent, absolute)){
			abstractFile = new File(absPath.substring(parentPath.length()+1, absPath.length()));
		}
		return abstractFile;
	}
	
	public static boolean isChild(File parent, File child){
		String parentPath = parent.getAbsolutePath().toLowerCase();
		String childPath = child.getAbsolutePath().toLowerCase();
		if(!childPath.startsWith(parentPath))
			return false;
		
		File f = child;
		while(f.getParentFile() != null){
			if(parent.equals(f))
				return true;
			f = f.getParentFile();
		}
		return false;
	}
}
